# -*- coding: utf-8 -*-
'''
A module for testing the logic of states and highstates

:codeauthor:    William Cannon <william.cannon@gmail.com>
:maturity:      new

Saltcheck provides unittest like functionality requiring only the knowledge of
salt module execution and yaml.

In order to run state and highstate saltcheck tests a sub-folder of a state must
be created and named ``saltcheck-tests``.

Tests for a state should be created in files ending in ``*.tst`` and placed in
the ``saltcheck-tests`` folder.

Multiple tests can be created in a file. Multiple ``*.tst`` files can be
created in the ``saltcheck-tests`` folder. Salt rendering is supported in test
files (e.g. ``yaml + jinja``). The ``id`` of a test works in the same manner as
in salt state files. They should be unique and descriptive.

Example file system layout:

.. code-block: txt

    /srv/salt/apache/
        init.sls
        config.sls
        saltcheck-tests/
            pkg_and_mods.tst
            config.tst

Example:

.. code-block:: yaml

    echo-test-hello:
      module_and_function: test.echo
      args:
        - "hello"
      kwargs:
      assertion: assertEqual
      expected-return:  'hello'

'''

# Import Python libs
from __future__ import absolute_import, unicode_literals, print_function
import logging
import os
import time
from json import loads, dumps

# Import Salt libs
import salt.utils.files
import salt.utils.path
import salt.utils.yaml
import salt.client
import salt.exceptions
from salt.ext import six

log = logging.getLogger(__name__)

__virtualname__ = 'saltcheck'


def __virtual__():
    '''
    Check dependencies - may be useful in future
    '''
    return __virtualname__


def update_master_cache():
    '''
    Updates the master cache onto the minion - transfers all salt-check-tests
    Should be done one time before running tests, and if tests are updated
    Can be automated by setting "auto_update_master_cache: True" in minion config

    CLI Example:

    .. code-block:: bash

        salt '*' saltcheck.update_master_cache
    '''
    __salt__['cp.cache_master']()
    return True


def run_test(**kwargs):
    '''
    Execute one saltcheck test and return result

    :param keyword arg test:

    CLI Example:

    .. code-block:: bash

        salt '*' saltcheck.run_test
            test='{"module_and_function": "test.echo",
                   "assertion": "assertEqual",
                   "expected-return": "This works!",
                   "args":["This works!"] }'
    '''
    # salt converts the string to a dictionary auto-magically
    scheck = SaltCheck()
    test = kwargs.get('test', None)
    if test and isinstance(test, dict):
        return scheck.run_test(test)
    else:
        return "Test must be a dictionary"


def run_state_tests(state):
    '''
    Execute all tests for a salt state and return results
    Nested states will also be tested

    :param str state: the name of a user defined state

    CLI Example:

    .. code-block:: bash

        salt '*' saltcheck.run_state_tests postfix
    '''
    scheck = SaltCheck()
    paths = scheck.get_state_search_path_list()
    stl = StateTestLoader(search_paths=paths)
    results = {}
    sls_list = _get_state_sls(state)
    for state_name in sls_list:
        mypath = stl.convert_sls_to_path(state_name)
        stl.add_test_files_for_sls(mypath)
        stl.load_test_suite()
        results_dict = {}
        for key, value in stl.test_dict.items():
            result = scheck.run_test(value)
            results_dict[key] = result
        results[state_name] = results_dict
    passed = 0
    failed = 0
    missing_tests = 0
    for state in results:
        if len(results[state].items()) == 0:
            missing_tests = missing_tests + 1
        else:
            for dummy, val in results[state].items():
                log.info("dummy=%s, val=%s", dummy, val)
                if val.startswith('Pass'):
                    passed = passed + 1
                if val.startswith('Fail'):
                    failed = failed + 1
    out_list = []
    for key, value in results.items():
        out_list.append({key: value})
    out_list.sort()
    out_list.append({"TEST RESULTS": {'Passed': passed, 'Failed': failed, 'Missing Tests': missing_tests}})
    return out_list


def run_highstate_tests():
    '''
    Execute all tests for a salt highstate and return results

    CLI Example:

    .. code-block:: bash

        salt '*' saltcheck.run_highstate_tests
    '''
    scheck = SaltCheck()
    paths = scheck.get_state_search_path_list()
    stl = StateTestLoader(search_paths=paths)
    results = {}
    sls_list = _get_top_states()
    all_states = []
    for top_state in sls_list:
        sls_list = _get_state_sls(top_state)
        for state in sls_list:
            if state not in all_states:
                all_states.append(state)

    for state_name in all_states:
        mypath = stl.convert_sls_to_path(state_name)
        stl.add_test_files_for_sls(mypath)
        stl.load_test_suite()
        results_dict = {}
        for key, value in stl.test_dict.items():
            result = scheck.run_test(value)
            results_dict[key] = result
        results[state_name] = results_dict
    passed = 0
    failed = 0
    missing_tests = 0
    for state in results:
        if len(results[state].items()) == 0:
            missing_tests = missing_tests + 1
        else:
            for dummy, val in results[state].items():
                log.info("dummy=%s, val=%s", dummy, val)
                if val.startswith('Pass'):
                    passed = passed + 1
                if val.startswith('Fail'):
                    failed = failed + 1
    out_list = []
    for key, value in results.items():
        out_list.append({key: value})
    out_list.sort()
    out_list.append({"TEST RESULTS": {'Passed': passed, 'Failed': failed, 'Missing Tests': missing_tests}})
    return out_list


def _render_file(file_path):
    '''
    call the salt utility to render a file
    '''
    # salt-call slsutil.renderer /srv/salt/jinjatest/saltcheck-tests/test1.tst
    rendered = __salt__['slsutil.renderer'](file_path)
    log.info("rendered: %s", rendered)
    return rendered


def _is_valid_module(module):
    '''
    Return a list of all modules available on minion
    '''
    modules = __salt__['sys.list_modules']()
    return bool(module in modules)


def _get_auto_update_cache_value():
    '''
    Return the config value of auto_update_master_cache
    '''
    __salt__['config.get']('auto_update_master_cache')
    return True


def _is_valid_function(module_name, function):
    '''
    Determine if a function is valid for a module
    '''
    try:
        functions = __salt__['sys.list_functions'](module_name)
    except salt.exceptions.SaltException:
        functions = ["unable to look up functions"]
    return "{0}.{1}".format(module_name, function) in functions


def _get_top_states():
    '''
    Equivalent to a salt cli: salt web state.show_top
    '''
    alt_states = []
    try:
        returned = __salt__['state.show_top']()
        for i in returned['base']:
            alt_states.append(i)
    except Exception:
        raise
    # log.info("top states: %s", alt_states)
    return alt_states


def _get_state_sls(state):
    '''
    Equivalent to a salt cli: salt web state.show_low_sls STATE
    '''
    sls_list_state = []
    try:
        returned = __salt__['state.show_low_sls'](state)
        for i in returned:
            if i['__sls__'] not in sls_list_state:
                sls_list_state.append(i['__sls__'])
    except Exception:
        raise
    return sls_list_state


class SaltCheck(object):
    '''
    This class implements the saltcheck
    '''

    def __init__(self):
        # self.sls_list_top = []
        self.sls_list_state = []
        self.modules = []
        self.results_dict = {}
        self.results_dict_summary = {}
        self.assertions_list = '''assertEqual assertNotEqual
                                  assertTrue assertFalse
                                  assertIn assertNotIn
                                  assertGreater
                                  assertGreaterEqual
                                  assertLess assertLessEqual'''.split()
        self.auto_update_master_cache = _get_auto_update_cache_value
        # self.salt_lc = salt.client.Caller(mopts=__opts__)
        self.salt_lc = salt.client.Caller()
        if self.auto_update_master_cache:
            update_master_cache()

    def __is_valid_test(self, test_dict):
        '''
        Determine if a test contains:

        - a test name
        - a valid module and function
        - a valid assertion
        - an expected return value
        '''
        tots = 0  # need total of >= 6 to be a valid test
        m_and_f = test_dict.get('module_and_function', None)
        assertion = test_dict.get('assertion', None)
        expected_return = test_dict.get('expected-return', None)
        log.info("__is_valid_test has test: %s", test_dict)
        if m_and_f:
            tots += 1
            module, function = m_and_f.split('.')
            if _is_valid_module(module):
                tots += 1
            if _is_valid_function(module, function):
                tots += 1
            log.info("__is_valid_test has valid m_and_f")
        if assertion:
            tots += 1
            if assertion in self.assertions_list:
                tots += 1
                log.info("__is_valid_test has valid_assertion")
        if expected_return:
            tots += 1
            log.info("__is_valid_test has valid_expected_return")
        log.info("__is_valid_test score: %s", tots)
        return tots >= 6

    def call_salt_command(self,
                          fun,
                          args,
                          kwargs):
        '''
        Generic call of salt Caller command
        '''
        value = False
        try:
            if args and kwargs:
                value = self.salt_lc.cmd(fun, *args, **kwargs)
            elif args and not kwargs:
                value = self.salt_lc.cmd(fun, *args)
            elif not args and kwargs:
                value = self.salt_lc.cmd(fun, **kwargs)
            else:
                value = self.salt_lc.cmd(fun)
        except salt.exceptions.SaltException:
            raise
        except Exception:
            raise
        return value

    def run_test(self, test_dict):
        '''
        Run a single saltcheck test
        '''
        if self.__is_valid_test(test_dict):
            mod_and_func = test_dict['module_and_function']
            args = test_dict.get('args', None)
            kwargs = test_dict.get('kwargs', None)
            assertion = test_dict['assertion']
            expected_return = test_dict['expected-return']
            actual_return = self.call_salt_command(mod_and_func, args, kwargs)
            if assertion != "assertIn":
                expected_return = self.cast_expected_to_returned_type(expected_return, actual_return)
            if assertion == "assertEqual":
                value = self.__assert_equal(expected_return, actual_return)
            elif assertion == "assertNotEqual":
                value = self.__assert_not_equal(expected_return, actual_return)
            elif assertion == "assertTrue":
                value = self.__assert_true(expected_return)
            elif assertion == "assertFalse":
                value = self.__assert_false(expected_return)
            elif assertion == "assertIn":
                value = self.__assert_in(expected_return, actual_return)
            elif assertion == "assertNotIn":
                value = self.__assert_not_in(expected_return, actual_return)
            elif assertion == "assertGreater":
                value = self.__assert_greater(expected_return, actual_return)
            elif assertion == "assertGreaterEqual":
                value = self.__assert_greater_equal(expected_return, actual_return)
            elif assertion == "assertLess":
                value = self.__assert_less(expected_return, actual_return)
            elif assertion == "assertLessEqual":
                value = self.__assert_less_equal(expected_return, actual_return)
            else:
                value = "Fail - bas assertion"
        else:
            return "Fail - invalid test"
        return value

    @staticmethod
    def cast_expected_to_returned_type(expected, returned):
        '''
        Determine the type of variable returned
        Cast the expected to the type of variable returned
        '''
        ret_type = type(returned)
        new_expected = expected
        if expected == "False" and ret_type == bool:
            expected = False
        try:
            new_expected = ret_type(expected)
        except ValueError:
            log.info("Unable to cast expected into type of returned")
            log.info("returned = %s", returned)
            log.info("type of returned = %s", type(returned))
            log.info("expected = %s", expected)
            log.info("type of expected = %s", type(expected))
        return new_expected

    @staticmethod
    def __assert_equal(expected, returned):
        '''
        Test if two objects are equal
        '''
        result = "Pass"

        try:
            assert (expected == returned), "{0} is not equal to {1}".format(expected, returned)
        except AssertionError as err:
            result = "Fail: " + six.text_type(err)
        return result

    @staticmethod
    def __assert_not_equal(expected, returned):
        '''
        Test if two objects are not equal
        '''
        result = "Pass"
        try:
            assert (expected != returned), "{0} is equal to {1}".format(expected, returned)
        except AssertionError as err:
            result = "Fail: " + six.text_type(err)
        return result

    @staticmethod
    def __assert_true(returned):
        '''
        Test if an boolean is True
        '''
        result = "Pass"
        try:
            assert (returned is True), "{0} not True".format(returned)
        except AssertionError as err:
            result = "Fail: " + six.text_type(err)
        return result

    @staticmethod
    def __assert_false(returned):
        '''
        Test if an boolean is False
        '''
        result = "Pass"
        if isinstance(returned, str):
            try:
                returned = bool(returned)
            except ValueError:
                raise
        try:
            assert (returned is False), "{0} not False".format(returned)
        except AssertionError as err:
            result = "Fail: " + six.text_type(err)
        return result

    @staticmethod
    def __assert_in(expected, returned):
        '''
        Test if a value is in the list of returned values
        '''
        result = "Pass"
        try:
            assert (expected in returned), "{0} not False".format(returned)
        except AssertionError as err:
            result = "Fail: " + six.text_type(err)
        return result

    @staticmethod
    def __assert_not_in(expected, returned):
        '''
        Test if a value is not in the list of returned values
        '''
        result = "Pass"
        try:
            assert (expected not in returned), "{0} not False".format(returned)
        except AssertionError as err:
            result = "Fail: " + six.text_type(err)
        return result

    @staticmethod
    def __assert_greater(expected, returned):
        '''
        Test if a value is greater than the returned value
        '''
        result = "Pass"
        try:
            assert (expected > returned), "{0} not False".format(returned)
        except AssertionError as err:
            result = "Fail: " + six.text_type(err)
        return result

    @staticmethod
    def __assert_greater_equal(expected, returned):
        '''
        Test if a value is greater than or equal to the returned value
        '''
        result = "Pass"
        try:
            assert (expected >= returned), "{0} not False".format(returned)
        except AssertionError as err:
            result = "Fail: " + six.text_type(err)
        return result

    @staticmethod
    def __assert_less(expected, returned):
        '''
        Test if a value is less than the returned value
        '''
        result = "Pass"
        try:
            assert (expected < returned), "{0} not False".format(returned)
        except AssertionError as err:
            result = "Fail: " + six.text_type(err)
        return result

    @staticmethod
    def __assert_less_equal(expected, returned):
        '''
        Test if a value is less than or equal to the returned value
        '''
        result = "Pass"
        try:
            assert (expected <= returned), "{0} not False".format(returned)
        except AssertionError as err:
            result = "Fail: " + six.text_type(err)
        return result

    @staticmethod
    def get_state_search_path_list():
        '''
        For the state file system, return a list of paths to search for states
        '''
        # state cache should be updated before running this method
        search_list = []
        cachedir = __opts__.get('cachedir', None)
        environment = __opts__['saltenv']
        if environment:
            path = cachedir + os.sep + "files" + os.sep + environment
            search_list.append(path)
        path = cachedir + os.sep + "files" + os.sep + "base"
        search_list.append(path)
        return search_list


class StateTestLoader(object):
    '''
    Class loads in test files for a state
    e.g. state_dir/saltcheck-tests/[1.tst, 2.tst, 3.tst]
    '''

    def __init__(self, search_paths):
        self.search_paths = search_paths
        self.path_type = None
        self.test_files = []  # list of file paths
        self.test_dict = {}

    def load_test_suite(self):
        '''
        Load tests either from one file, or a set of files
        '''
        self.test_dict = {}
        for myfile in self.test_files:
            # self.load_file(myfile)
            self.load_file_salt_rendered(myfile)
        self.test_files = []

    def load_file(self, filepath):
        '''
        loads in one test file
        '''
        try:
            with __utils__['files.fopen'](filepath, 'r') as myfile:
                # with salt.utils.files.fopen(filepath, 'r') as myfile:
                # with open(filepath, 'r') as myfile:
                contents_yaml = salt.utils.data.decode(salt.utils.yaml.safe_load(myfile))
                for key, value in contents_yaml.items():
                    self.test_dict[key] = value
        except:
            raise
        return

    def load_file_salt_rendered(self, filepath):
        '''
        loads in one test file
        '''
        # use the salt renderer module to interpret jinja and etc
        tests = _render_file(filepath)
        # use json as a convenient way to convert the OrderedDicts from salt renderer
        mydict = loads(dumps(tests))
        for key, value in mydict.items():
            self.test_dict[key] = value
        return

    def gather_files(self, filepath):
        '''
        Gather files for a test suite
        '''
        self.test_files = []
        log.info("gather_files: %s", time.time())
        filepath = filepath + os.sep + 'saltcheck-tests'
        rootdir = filepath
        # for dirname, subdirlist, filelist in salt.utils.path.os_walk(rootdir):
        for dirname, dummy, filelist in salt.utils.path.os_walk(rootdir):
            for fname in filelist:
                if fname.endswith('.tst'):
                    start_path = dirname + os.sep + fname
                    full_path = os.path.abspath(start_path)
                    self.test_files.append(full_path)
        return

    @staticmethod
    def convert_sls_to_paths(sls_list):
        '''
        Converting sls to paths
        '''
        new_sls_list = []
        for sls in sls_list:
            sls = sls.replace(".", os.sep)
            new_sls_list.append(sls)
        return new_sls_list

    @staticmethod
    def convert_sls_to_path(sls):
        '''
        Converting sls to paths
        '''
        sls = sls.replace(".", os.sep)
        return sls

    def add_test_files_for_sls(self, sls_path):
        '''
        Adding test files
        '''
        for path in self.search_paths:
            full_path = path + os.sep + sls_path
            rootdir = full_path
            if os.path.isdir(full_path):
                log.info("searching path= %s", full_path)
                # for dirname, subdirlist, filelist in salt.utils.path.os_walk(rootdir, topdown=True):
                for dirname, subdirlist, dummy in salt.utils.path.os_walk(rootdir, topdown=True):
                    if "saltcheck-tests" in subdirlist:
                        self.gather_files(dirname)
                        log.info("test_files list: %s", self.test_files)
                        log.info("found subdir match in = %s", dirname)
                    else:
                        log.info("did not find subdir match in = %s", dirname)
                    del subdirlist[:]
            else:
                log.info("path is not a directory= %s", full_path)
        return



# -*- coding: utf-8 -*-
'''
Opening of Encrypted Devices
=======================

Ensure that an encrypted device is mapped with the `mapped` function:

.. code-block:: yaml

    mappedname:
      cryptdev.mapped:
        - device: /dev/sdb1
        - keyfile: /etc/keyfile.key
        - opts:
          - size=256

    swap:
      crypted.mapped:
        - device: /dev/sdx4
        - keyfile: /dev/urandom
        - opts: swap,cipher=aes-cbc-essiv:sha256,size=256

    mappedbyuuid:
      crypted.mapped:
        - device: UUID=066e0200-2867-4ebe-b9e6-f30026ca2314
        - keyfile: /etc/keyfile.key
        - config: /etc/alternate-crypttab

.. versionadded:: 2018.3.0
'''
from __future__ import absolute_import, print_function, unicode_literals

import logging
log = logging.getLogger(__name__)


def mapped(name,
           device,
           keyfile=None,
           opts=None,
           config='/etc/crypttab',
           persist=True,
           immediate=False,
           match_on='name'):
    '''
    Verify that a device is mapped

    name
        The name under which the device is to be mapped

    device
        The device name, typically the device node, such as ``/dev/sdb1``
        or ``UUID=066e0200-2867-4ebe-b9e6-f30026ca2314``.

    keyfile
        Either ``None`` if the password is to be entered manually on boot, or
        an absolute path to a keyfile. If the password is to be asked
        interactively, the mapping cannot be performed with ``immediate=True``.

    opts
        A list object of options or a comma delimited list

    config
        Set an alternative location for the crypttab, if the map is persistent,
        Default is ``/etc/crypttab``

    persist
        Set if the map should be saved in the crypttab, Default is ``True``

    immediate
        Set if the device mapping should be executed immediately. Requires that
        the keyfile not be ``None``, because the password cannot be asked
        interactively. Note that options are not passed through on the initial
        mapping.  Default is ``False``.

    match_on
        A name or list of crypttab properties on which this state should be applied.
        Default is ``name``, meaning that the line is matched only by the name
        parameter. If the desired configuration requires two devices mapped to
        the same name, supply a list of parameters to match on.
    '''
    ret = {'name': name,
           'changes': {},
           'result': True,
           'comment': ''}

    # If neither option is set, we've been asked to do nothing.
    if not immediate and not persist:
        ret['result'] = False
        ret['comment'] = 'Either persist or immediate must be set, otherwise this state does nothing'
        return ret

    if immediate and (keyfile is None or keyfile == 'none' or keyfile == '-'):
        ret['result'] = False
        ret['changes']['cryptsetup'] = 'Device cannot be mapped immediately without a keyfile'

    elif immediate:
        # Get the active crypt mounts. If ours is listed already, no action is necessary.
        active = __salt__['cryptdev.active']()
        if name not in active.keys():
            # Open the map using cryptsetup. This does not pass any options.
            if opts:
                log.warn('passed cryptdev options are ignored when mapping immediately')

            if __opts__['test']:
                ret['result'] = None
                ret['commment'] = 'Device would be mapped immediately'
            else:
                cryptsetup_result = __salt__['cryptdev.open'](name, device, keyfile)
                if cryptsetup_result:
                    ret['changes']['cryptsetup'] = 'Device mapped using cryptsetup'
                else:
                    ret['changes']['cryptsetup'] = 'Device failed to map using cryptsetup'
                    ret['result'] = False

    if persist and not __opts__['test']:
        crypttab_result = __salt__['cryptdev.set_crypttab'](name,
                                                            device,
                                                            password=keyfile,
                                                            options=opts,
                                                            config=config,
                                                            match_on=match_on)
        if crypttab_result:
            if crypttab_result == 'new':
                ret['changes']['crypttab'] = 'Entry added in {0}'.format(config)

            if crypttab_result == 'change':
                ret['changes']['crypttab'] = 'Existing entry in {0} changed'.format(config)

        else:
            ret['changes']['crypttab'] = 'Unable to set entry in {0}'.format(config)
            ret['result'] = False

    return ret


def unmapped(name,
             config='/etc/crypttab',
             persist=True,
             immediate=False):
    '''
    Ensure that a device is unmapped

    name
        The name to ensure is not mapped

    config
        Set an alternative location for the crypttab, if the map is persistent,
        Default is ``/etc/crypttab``

    persist
        Set if the map should be removed from the crypttab. Default is ``True``

    immediate
        Set if the device should be unmapped immediately. Default is ``False``.
    '''
    ret = {'name': name,
           'changes': {},
           'result': True,
           'comment': ''}

    if immediate:
        # Get the active crypt mounts. If ours is not listed already, no action is necessary.
        active = __salt__['cryptdev.active']()
        if name in active.keys():
            # Close the map using cryptsetup.
            if __opts__['test']:
                ret['result'] = None
                ret['commment'] = 'Device would be unmapped immediately'
            else:
                cryptsetup_result = __salt__['cryptdev.close'](name)
                if cryptsetup_result:
                    ret['changes']['cryptsetup'] = 'Device unmapped using cryptsetup'
                else:
                    ret['changes']['cryptsetup'] = 'Device failed to unmap using cryptsetup'
                    ret['result'] = False

    if persist and not __opts__['test']:
        crypttab_result = __salt__['cryptdev.rm_crypttab'](name, config=config)
        if crypttab_result:
            if crypttab_result == 'change':
                ret['changes']['crypttab'] = 'Entry removed from {0}'.format(config)

        else:
            ret['changes']['crypttab'] = 'Unable to remove entry in {0}'.format(config)
            ret['result'] = False

    return ret



# -*- coding: utf-8 -*-
'''
    Log4Mongo Logging Handler
    =========================

    This module provides a logging handler for sending salt logs to MongoDB

    Configuration
    -------------

    In the salt configuration file (e.g. /etc/salt/{master,minion}):

    .. code-block:: yaml

        log4mongo_handler:
          host: mongodb_host
          port: 27017
          database_name: logs
          collection: salt_logs
          username: logging
          password: reindeerflotilla
          write_concern: 0
          log_level: warning


    Log Level
    .........

    If not set, the log_level will be set to the level defined in the global
    configuration file setting.

    .. admonition:: Inspiration

        This work was inspired by the Salt logging handlers for LogStash and
        Sentry and by the log4mongo Python implementation.
'''
# Import python libs
from __future__ import absolute_import, print_function, unicode_literals
import socket
import logging

# Import salt libs
from salt.ext import six
from salt.log.mixins import NewStyleClassMixIn
from salt.log.setup import LOG_LEVELS

# Import third party libs
try:
    from log4mongo.handlers import MongoHandler, MongoFormatter
    HAS_MONGO = True
except ImportError:
    HAS_MONGO = False

__virtualname__ = 'mongo'


def __virtual__():
    if not HAS_MONGO:
        return False
    return __virtualname__


class FormatterWithHost(logging.Formatter, NewStyleClassMixIn):
    def format(self, record):
        mongoformatter = MongoFormatter()
        document = mongoformatter.format(record)
        document['hostname'] = socket.gethostname()
        return document


def setup_handlers():
    handler_id = 'log4mongo_handler'
    if handler_id in __opts__:
        config_fields = {
            'host': 'host',
            'port': 'port',
            'database_name': 'database_name',
            'collection': 'collection',
            'username': 'username',
            'password': 'password',
            'write_concern': 'w'
        }

        config_opts = {}
        for config_opt, arg_name in six.iteritems(config_fields):
            config_opts[arg_name] = __opts__[handler_id].get(config_opt)

        config_opts['level'] = LOG_LEVELS[
            __opts__[handler_id].get(
                'log_level',
                __opts__.get('log_level', 'error')
            )
        ]

        handler = MongoHandler(
            formatter=FormatterWithHost(),
            **config_opts
        )
        yield handler
    else:
        yield False



# -*- coding: utf-8 -*-
'''
Management of Zabbix mediatypes.

:codeauthor: Raymond Kuiper <qix@the-wired.net>

'''

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt libs
from salt.ext import six


def __virtual__():
    '''
    Only make these states available if Zabbix module is available.
    '''
    return 'zabbix.mediatype_create' in __salt__


def present(name, mediatype, **kwargs):
    '''
    Creates new mediatype.
    NOTE: This function accepts all standard mediatype properties: keyword argument names differ depending on your
    zabbix version, see:
    https://www.zabbix.com/documentation/3.0/manual/api/reference/host/object#host_inventory

    :param name: name of the mediatype
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)

    .. code-block:: yaml

        make_new_mediatype:
            zabbix_mediatype.present:
                - name: 'Email'
                - mediatype: 0
                - smtp_server: smtp.example.com
                - smtp_hello: zabbix.example.com
                - smtp_email: zabbix@example.com

    '''
    connection_args = {}
    if '_connection_user' in kwargs:
        connection_args['_connection_user'] = kwargs['_connection_user']
    if '_connection_password' in kwargs:
        connection_args['_connection_password'] = kwargs['_connection_password']
    if '_connection_url' in kwargs:
        connection_args['_connection_url'] = kwargs['_connection_url']

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    # Comment and change messages
    comment_mediatype_created = 'Mediatype {0} created.'.format(name)
    comment_mediatype_updated = 'Mediatype {0} updated.'.format(name)
    comment_mediatype_notcreated = 'Unable to create mediatype: {0}. '.format(name)
    comment_mediatype_exists = 'Mediatype {0} already exists.'.format(name)
    changes_mediatype_created = {name: {'old': 'Mediatype {0} does not exist.'.format(name),
                                        'new': 'Mediatype {0} created.'.format(name),
                                        }
                                 }

    # Zabbix API expects script parameters as a string of arguments seperated by newline characters
    if 'exec_params' in kwargs:
        if isinstance(kwargs['exec_params'], list):
            kwargs['exec_params'] = '\n'.join(kwargs['exec_params'])+'\n'
        else:
            kwargs['exec_params'] = six.text_type(kwargs['exec_params'])+'\n'

    mediatype_exists = __salt__['zabbix.mediatype_get'](name, **connection_args)

    if mediatype_exists:
        mediatypeobj = mediatype_exists[0]
        mediatypeid = int(mediatypeobj['mediatypeid'])
        update_email = False
        update_email_port = False
        update_email_security = False
        update_email_verify_peer = False
        update_email_verify_host = False
        update_email_auth = False
        update_script = False
        update_script_params = False
        update_sms = False
        update_jabber = False
        update_eztext = False
        update_status = False

        if int(mediatype) == 0 and 'smtp_server' in kwargs and 'smtp_helo' in kwargs and 'smtp_email' in kwargs:
            if (int(mediatype) != int(mediatypeobj['type']) or
                    kwargs['smtp_server'] != mediatypeobj['smtp_server'] or
                    kwargs['smtp_email'] != mediatypeobj['smtp_email'] or
                    kwargs['smtp_helo'] != mediatypeobj['smtp_helo']):
                update_email = True

        if int(mediatype) == 0 and 'smtp_port' in kwargs:
            if int(kwargs['smtp_port']) != int(mediatypeobj['smtp_port']):
                update_email_port = True

        if int(mediatype) == 0 and 'smtp_security' in kwargs:
            if int(kwargs['smtp_security']) != int(mediatypeobj['smtp_security']):
                update_email_security = True

        if int(mediatype) == 0 and 'smtp_verify_peer' in kwargs:
            if int(kwargs['smtp_verify_peer']) != int(mediatypeobj['smtp_verify_peer']):
                update_email_verify_peer = True

        if int(mediatype) == 0 and 'smtp_verify_host' in kwargs:
            if int(kwargs['smtp_verify_host']) != int(mediatypeobj['smtp_verify_host']):
                update_email_verify_host = True

        if int(mediatype) == 0 and 'smtp_authentication' in kwargs and 'username' in kwargs and 'passwd' in kwargs:
            if (int(kwargs['smtp_authentication']) != int(mediatypeobj['smtp_authentication']) or
                    kwargs['username'] != mediatypeobj['username'] or
                    kwargs['passwd'] != mediatypeobj['passwd']):
                update_email_auth = True

        if int(mediatype) == 1 and 'exec_path' in kwargs:
            if (int(mediatype) != int(mediatypeobj['type']) or
                    kwargs['exec_path'] != mediatypeobj['exec_path']):
                update_script = True

        if int(mediatype) == 1 and 'exec_params' in kwargs:
            if kwargs['exec_params'] != mediatypeobj['exec_params']:
                update_script_params = True

        if int(mediatype) == 2 and 'gsm_modem' in kwargs:
            if (int(mediatype) != int(mediatypeobj['type']) or
                    kwargs['gsm_modem'] != mediatypeobj['gsm_modem']):
                update_sms = True

        if int(mediatype) == 3 and 'username' in kwargs and 'passwd' in kwargs:
            if (int(mediatype) != int(mediatypeobj['type']) or
                    kwargs['username'] != mediatypeobj['username'] or
                    kwargs['passwd'] != mediatypeobj['passwd']):
                update_jabber = True

        if int(mediatype) == 100 and 'username' in kwargs and 'passwd' in kwargs and 'exec_path' in kwargs:
            if (int(mediatype) != int(mediatypeobj['type']) or
                    kwargs['username'] != mediatypeobj['username'] or
                    kwargs['passwd'] != mediatypeobj['passwd'] or
                    kwargs['exec_path'] != mediatypeobj['exec_path']):
                update_eztext = True

        if 'status' in kwargs:
            if int(kwargs['status']) != int(mediatypeobj['status']):
                update_status = True

    # Dry run, test=true mode
    if __opts__['test']:
        if mediatype_exists:
            if update_status:
                ret['result'] = None
                ret['comment'] = comment_mediatype_updated
            else:
                ret['result'] = True
                ret['comment'] = comment_mediatype_exists
        else:
            ret['result'] = None
            ret['comment'] = comment_mediatype_created
        return ret

    error = []

    if mediatype_exists:
        if (update_email or update_email_port or update_email_security or
                update_email_verify_peer or update_email_verify_host or update_email_auth or
                update_script or update_script_params or update_sms or
                update_jabber or update_eztext or update_status):
            ret['result'] = True
            ret['comment'] = comment_mediatype_updated

            if update_email:
                updated_email = __salt__['zabbix.mediatype_update'](mediatypeid,
                                                                    type=mediatype,
                                                                    smtp_server=kwargs['smtp_server'],
                                                                    smtp_helo=kwargs['smtp_helo'],
                                                                    smtp_email=kwargs['smtp_email'],
                                                                    **connection_args)
                if 'error' in updated_email:
                    error.append(updated_email['error'])
                else:
                    ret['changes']['smtp_server'] = kwargs['smtp_server']
                    ret['changes']['smtp_helo'] = kwargs['smtp_helo']
                    ret['changes']['smtp_email'] = kwargs['smtp_email']

            if update_email_port:
                updated_email_port = __salt__['zabbix.mediatype_update'](mediatypeid,
                                                                         smtp_port=kwargs['smtp_port'],
                                                                         **connection_args)
                if 'error' in updated_email_port:
                    error.append(updated_email_port['error'])
                else:
                    ret['changes']['smtp_port'] = kwargs['smtp_port']

            if update_email_security:
                updated_email_security = __salt__['zabbix.mediatype_update'](mediatypeid,
                                                                             smtp_security=kwargs['smtp_security'],
                                                                             **connection_args)
                if 'error' in updated_email_security:
                    error.append(updated_email_security['error'])
                else:
                    ret['changes']['smtp_security'] = kwargs['smtp_security']

            if update_email_verify_peer:
                updated_email_verify_peer = __salt__['zabbix.mediatype_update'](mediatypeid,
                                                                                smtp_verify_peer=kwargs['smtp_verify_peer'],
                                                                                **connection_args)
                if 'error' in updated_email_verify_peer:
                    error.append(updated_email_verify_peer['error'])
                else:
                    ret['changes']['smtp_verify_peer'] = kwargs['smtp_verify_peer']

            if update_email_verify_host:
                updated_email_verify_host = __salt__['zabbix.mediatype_update'](mediatypeid,
                                                                                smtp_verify_host=kwargs['smtp_verify_host'],
                                                                                **connection_args)
                if 'error' in updated_email_verify_host:
                    error.append(updated_email_verify_host['error'])
                else:
                    ret['changes']['smtp_verify_host'] = kwargs['smtp_verify_host']

            if update_email_auth:
                updated_email_auth = __salt__['zabbix.mediatype_update'](mediatypeid,
                                                                         username=kwargs['username'],
                                                                         passwd=kwargs['passwd'],
                                                                         smtp_authentication=kwargs['smtp_authentication'],
                                                                         **connection_args)
                if 'error' in updated_email_auth:
                    error.append(updated_email_auth['error'])
                else:
                    ret['changes']['smtp_authentication'] = kwargs['smtp_authentication']
                    ret['changes']['username'] = kwargs['username']

            if update_script:
                updated_script = __salt__['zabbix.mediatype_update'](mediatypeid,
                                                                     type=mediatype,
                                                                     exec_path=kwargs['exec_path'],
                                                                     **connection_args)
                if 'error' in updated_script:
                    error.append(updated_script['error'])
                else:
                    ret['changes']['exec_path'] = kwargs['exec_path']

            if update_script_params:
                updated_script_params = __salt__['zabbix.mediatype_update'](mediatypeid,
                                                                            exec_params=kwargs['exec_params'],
                                                                            **connection_args)
                if 'error' in updated_script_params:
                    error.append(updated_script['error'])
                else:
                    ret['changes']['exec_params'] = kwargs['exec_params']

            if update_sms:
                updated_sms = __salt__['zabbix.mediatype_update'](mediatypeid,
                                                                  type=mediatype,
                                                                  gsm_modem=kwargs['gsm_modem'],
                                                                  **connection_args)
                if 'error' in updated_sms:
                    error.append(updated_sms['error'])
                else:
                    ret['changes']['gsm_modem'] = kwargs['gsm_modem']

            if update_jabber:
                updated_jabber = __salt__['zabbix.mediatype_update'](mediatypeid,
                                                                     type=mediatype,
                                                                     username=kwargs['username'],
                                                                     passwd=kwargs['passwd'],
                                                                     **connection_args)
                if 'error' in updated_jabber:
                    error.append(updated_jabber['error'])
                else:
                    ret['changes']['username'] = kwargs['username']

            if update_eztext:
                updated_eztext = __salt__['zabbix.mediatype_update'](mediatypeid,
                                                                     type=mediatype,
                                                                     username=kwargs['username'],
                                                                     passwd=kwargs['passwd'],
                                                                     exec_path=kwargs['exec_path'],
                                                                     **connection_args)
                if 'error' in updated_eztext:
                    error.append(updated_eztext['error'])
                else:
                    ret['changes']['username'] = kwargs['username']
                    ret['changes']['exec_path'] = kwargs['exec_path']

            if update_status:
                updated_status = __salt__['zabbix.mediatype_update'](mediatypeid,
                                                                     status=kwargs['status'],
                                                                     **connection_args)
                if 'error' in updated_status:
                    error.append(updated_status['error'])
                else:
                    ret['changes']['status'] = kwargs['status']

        else:
            ret['result'] = True
            ret['comment'] = comment_mediatype_exists
    else:
        mediatype_create = __salt__['zabbix.mediatype_create'](name, mediatype, **kwargs)

        if 'error' not in mediatype_create:
            ret['result'] = True
            ret['comment'] = comment_mediatype_created
            ret['changes'] = changes_mediatype_created
        else:
            ret['result'] = False
            ret['comment'] = comment_mediatype_notcreated + six.text_type(mediatype_create['error'])

    # error detected
    if error:
        ret['changes'] = {}
        ret['result'] = False
        ret['comment'] = six.text_type(error)

    return ret


def absent(name, **kwargs):
    '''
    Ensures that the mediatype does not exist, eventually deletes the mediatype.

    :param name: name of the mediatype
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)

    .. code-block:: yaml

        delete_mediatype:
            zabbix_mediatype.absent:
                - name: 'Email'
    '''
    connection_args = {}
    if '_connection_user' in kwargs:
        connection_args['_connection_user'] = kwargs['_connection_user']
    if '_connection_password' in kwargs:
        connection_args['_connection_password'] = kwargs['_connection_password']
    if '_connection_url' in kwargs:
        connection_args['_connection_url'] = kwargs['_connection_url']

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    # Comment and change messages
    comment_mediatype_deleted = 'Mediatype {0} deleted.'.format(name)
    comment_mediatype_notdeleted = 'Unable to delete mediatype: {0}. '.format(name)
    comment_mediatype_notexists = 'Mediatype {0} does not exist.'.format(name)
    changes_mediatype_deleted = {name: {'old': 'Mediatype {0} exists.'.format(name),
                                        'new': 'Mediatype {0} deleted.'.format(name),
                                        }
                                 }

    mediatype_exists = __salt__['zabbix.mediatype_get'](name, **connection_args)

    # Dry run, test=true mode
    if __opts__['test']:
        if not mediatype_exists:
            ret['result'] = True
            ret['comment'] = comment_mediatype_notexists
        else:
            ret['result'] = None
            ret['comment'] = comment_mediatype_deleted
        return ret

    if not mediatype_exists:
        ret['result'] = True
        ret['comment'] = comment_mediatype_notexists
    else:
        try:
            mediatypeid = mediatype_exists[0]['mediatypeid']
            mediatype_delete = __salt__['zabbix.mediatype_delete'](mediatypeid, **connection_args)
        except KeyError:
            mediatype_delete = False

        if mediatype_delete and 'error' not in mediatype_delete:
            ret['result'] = True
            ret['comment'] = comment_mediatype_deleted
            ret['changes'] = changes_mediatype_deleted
        else:
            ret['result'] = False
            ret['comment'] = comment_mediatype_notdeleted + six.text_type(mediatype_delete['error'])

    return ret



# -*- coding: utf-8 -*-
'''
Configure a Pacemaker/Corosync cluster with PCS
===============================================

Configure Pacemaker/Cororsync clusters with the
Pacemaker/Cororsync conifguration system (PCS)

:depends: pcs

.. versionadded:: 2016.3.0
'''
from __future__ import absolute_import, print_function, unicode_literals

# Import salt libs
import salt.utils.path
from salt.ext import six


def __virtual__():
    '''
    Only load if pcs package is installed
    '''
    if salt.utils.path.which('pcs'):
        return 'pcs'
    return False


def item_show(item, item_id=None, item_type=None, show='show', extra_args=None, cibfile=None):
    '''
    Show an item via pcs command
    (mainly for use with the pcs state module)

    item
        config, property, resource, constraint etc.
    item_id
        id of the item
    item_type
        item type
    show
        show command (probably None, default: show)
    extra_args
        additional options for the pcs command
    cibfile
        use cibfile instead of the live CIB
    '''
    cmd = ['pcs']

    if isinstance(cibfile, six.string_types):
        cmd += ['-f', cibfile]

    if isinstance(item, six.string_types):
        cmd += [item]
    elif isinstance(item, (list, tuple)):
        cmd += item

    # constraint command follows a different order
    if item in ['constraint']:
        cmd += [item_type]

    if isinstance(show, six.string_types):
        cmd += [show]
    elif isinstance(show, (list, tuple)):
        cmd += show

    if isinstance(item_id, six.string_types):
        cmd += [item_id]

    if isinstance(extra_args, (list, tuple)):
        cmd += extra_args

    # constraint command only shows id, when using '--full'-parameter
    if item in ['constraint']:
        if not isinstance(extra_args, (list, tuple)) or '--full' not in extra_args:
            cmd += ['--full']

    return __salt__['cmd.run_all'](cmd, output_loglevel='trace', python_shell=False)


def item_create(item, item_id, item_type, create='create', extra_args=None, cibfile=None):
    '''
    Create an item via pcs command
    (mainly for use with the pcs state module)

    item
        config, property, resource, constraint etc.
    item_id
        id of the item
    item_type
        item type
    create
        create command (create or set f.e., default: create)
    extra_args
        additional options for the pcs command
    cibfile
        use cibfile instead of the live CIB
    '''
    cmd = ['pcs']
    if isinstance(cibfile, six.string_types):
        cmd += ['-f', cibfile]

    if isinstance(item, six.string_types):
        cmd += [item]
    elif isinstance(item, (list, tuple)):
        cmd += item

    # constraint command follows a different order
    if item in ['constraint']:
        if isinstance(item_type, six.string_types):
            cmd += [item_type]

    if isinstance(create, six.string_types):
        cmd += [create]
    elif isinstance(create, (list, tuple)):
        cmd += create

    # constraint command needs item_id in format 'id=<id' after all params
    # constraint command follows a different order
    if item not in ['constraint']:
        cmd += [item_id]
        if isinstance(item_type, six.string_types):
            cmd += [item_type]

    if isinstance(extra_args, (list, tuple)):
        # constraint command needs item_id in format 'id=<id' after all params
        if item in ['constraint']:
            extra_args = extra_args + ['id={0}'.format(item_id)]
        cmd += extra_args

    return __salt__['cmd.run_all'](cmd, output_loglevel='trace', python_shell=False)


def auth(nodes, pcsuser='hacluster', pcspasswd='hacluster', extra_args=None):
    '''
    Authorize nodes to the cluster

    nodes
        a list of nodes which should be authorized to the cluster
    pcsuser
        user for communitcation with PCS (default: hacluster)
    pcspasswd
        password for pcsuser (default: hacluster)
    extra_args
        list of extra option for the \'pcs cluster auth\' command

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.auth nodes='[ node1.example.org node2.example.org ]' pcsuser=hacluster pcspasswd=hoonetorg extra_args="[ '--force' ]"
    '''
    cmd = ['pcs', 'cluster', 'auth']

    if pcsuser:
        cmd += ['-u', pcsuser]

    if pcspasswd:
        cmd += ['-p', pcspasswd]

    if isinstance(extra_args, (list, tuple)):
        cmd += extra_args
    cmd += nodes

    return __salt__['cmd.run_all'](cmd, output_loglevel='trace', python_shell=False)


def is_auth(nodes):
    '''
    Check if nodes are already authorized

    nodes
        a list of nodes to be checked for authorization to the cluster

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.is_auth nodes='[node1.example.org node2.example.org]'
    '''
    cmd = ['pcs', 'cluster', 'auth']
    cmd += nodes

    return __salt__['cmd.run_all'](cmd, stdin='\n\n', output_loglevel='trace', python_shell=False)


def cluster_setup(nodes, pcsclustername='pcscluster', extra_args=None):
    '''
    Setup pacemaker cluster via pcs command

    nodes
        a list of nodes which should be set up
    pcsclustername
        Name of the Pacemaker cluster (default: pcscluster)
    extra_args
        list of extra option for the \'pcs cluster setup\' command

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.cluster_setup nodes='[ node1.example.org node2.example.org ]' pcsclustername=pcscluster
    '''
    cmd = ['pcs', 'cluster', 'setup']

    cmd += ['--name', pcsclustername]

    cmd += nodes
    if isinstance(extra_args, (list, tuple)):
        cmd += extra_args

    return __salt__['cmd.run_all'](cmd, output_loglevel='trace', python_shell=False)


def cluster_node_add(node, extra_args=None):
    '''
    Add a node to the pacemaker cluster via pcs command

    node
        node that should be added
    extra_args
        list of extra option for the \'pcs cluster node add\' command

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.cluster_node_add node=node2.example.org
    '''
    cmd = ['pcs', 'cluster', 'node', 'add']

    cmd += [node]
    if isinstance(extra_args, (list, tuple)):
        cmd += extra_args

    return __salt__['cmd.run_all'](cmd, output_loglevel='trace', python_shell=False)


def cib_create(cibfile, scope='configuration', extra_args=None):
    '''
    Create a CIB-file from the current CIB of the cluster

    cibfile
        name/path of the file containing the CIB
    scope
        specific section of the CIB (default: configuration)
    extra_args
        additional options for creating the CIB-file

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.cib_create cibfile='/tmp/VIP_apache_1.cib' scope=False
    '''
    cmd = ['pcs', 'cluster', 'cib', cibfile]
    if isinstance(scope, six.string_types):
        cmd += ['scope={0}'.format(scope)]
    if isinstance(extra_args, (list, tuple)):
        cmd += extra_args

    return __salt__['cmd.run_all'](cmd, output_loglevel='trace', python_shell=False)


def cib_push(cibfile, scope='configuration', extra_args=None):
    '''
    Push a CIB-file as the new CIB to the cluster

    cibfile
        name/path of the file containing the CIB
    scope
        specific section of the CIB (default: configuration)
    extra_args
        additional options for creating the CIB-file

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.cib_push cibfile='/tmp/VIP_apache_1.cib' scope=False
    '''
    cmd = ['pcs', 'cluster', 'cib-push', cibfile]
    if isinstance(scope, six.string_types):
        cmd += ['scope={0}'.format(scope)]
    if isinstance(extra_args, (list, tuple)):
        cmd += extra_args

    return __salt__['cmd.run_all'](cmd, output_loglevel='trace', python_shell=False)


def config_show(cibfile=None):
    '''
    Show config of cluster

    cibfile
        name/path of the file containing the CIB

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.config_show cibfile='/tmp/cib_for_galera'
    '''
    return item_show(item='config', item_id=None, extra_args=None, cibfile=cibfile)


def prop_show(prop, extra_args=None, cibfile=None):
    '''
    Show the value of a cluster property

    prop
        name of the property
    extra_args
        additional options for the pcs property command
    cibfile
        use cibfile instead of the live CIB

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.prop_show cibfile='/tmp/2_node_cluster.cib' prop='no-quorum-policy' cibfile='/tmp/2_node_cluster.cib'
    '''
    return item_show(item='property', item_id=prop, extra_args=extra_args, cibfile=cibfile)


def prop_set(prop, value, extra_args=None, cibfile=None):
    '''
    Set the value of a cluster property

    prop
        name of the property
    value
        value of the property prop
    extra_args
        additional options for the pcs property command
    cibfile
        use cibfile instead of the live CIB

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.prop_set prop='no-quorum-policy' value='ignore' cibfile='/tmp/2_node_cluster.cib'
    '''
    return item_create(item='property',
                       item_id='{0}={1}'.format(prop, value),
                       item_type=None,
                       create='set',
                       extra_args=extra_args,
                       cibfile=cibfile)


def stonith_show(stonith_id, extra_args=None, cibfile=None):
    '''
    Show the value of a cluster stonith

    stonith_id
        name for the stonith resource
    extra_args
        additional options for the pcs stonith command
    cibfile
        use cibfile instead of the live CIB

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.stonith_show stonith_id='eps_fence' cibfile='/tmp/2_node_cluster.cib'
    '''
    return item_show(item='stonith', item_id=stonith_id, extra_args=extra_args, cibfile=cibfile)


def stonith_create(stonith_id, stonith_device_type, stonith_device_options=None, cibfile=None):
    '''
    Create a stonith resource via pcs command

    stonith_id
        name for the stonith resource
    stonith_device_type
        name of the stonith agent fence_eps, fence_xvm f.e.
    stonith_device_options
        additional options for creating the stonith resource
    cibfile
        use cibfile instead of the live CIB for manipulation

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.stonith_create stonith_id='eps_fence' stonith_device_type='fence_eps'
                                    stonith_device_options="['pcmk_host_map=node1.example.org:01;node2.example.org:02', 'ipaddr=myepsdevice.example.org', 'action=reboot', 'power_wait=5', 'verbose=1', 'debug=/var/log/pcsd/eps_fence.log', 'login=hidden', 'passwd=hoonetorg']" cibfile='/tmp/cib_for_stonith.cib'
    '''
    return item_create(item='stonith',
                       item_id=stonith_id,
                       item_type=stonith_device_type,
                       extra_args=stonith_device_options,
                       cibfile=cibfile)


def resource_show(resource_id, extra_args=None, cibfile=None):
    '''
    Show a resource via pcs command

    resource_id
        name of the resource
    extra_args
        additional options for the pcs command
    cibfile
        use cibfile instead of the live CIB

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.resource_show resource_id='galera' cibfile='/tmp/cib_for_galera.cib'
    '''
    return item_show(item='resource', item_id=resource_id, extra_args=extra_args, cibfile=cibfile)


def resource_create(resource_id, resource_type, resource_options=None, cibfile=None):
    '''
    Create a resource via pcs command

    resource_id
        name for the resource
    resource_type
        resource type (f.e. ocf:heartbeat:IPaddr2 or VirtualIP)
    resource_options
        additional options for creating the resource
    cibfile
        use cibfile instead of the live CIB for manipulation

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.resource_create resource_id='galera' resource_type='ocf:heartbeat:galera' resource_options="['wsrep_cluster_address=gcomm://node1.example.org,node2.example.org,node3.example.org', '--master']" cibfile='/tmp/cib_for_galera.cib'
    '''
    return item_create(item='resource',
                       item_id=resource_id,
                       item_type=resource_type,
                       extra_args=resource_options,
                       cibfile=cibfile)



# -*- coding: utf-8 -*-
'''
System Profiler Module

Interface with macOS's command-line System Profiler utility to get
information about package receipts and installed applications.

.. versionadded:: 2015.5.0

'''

from __future__ import absolute_import, unicode_literals, print_function

import plistlib
import subprocess
import salt.utils.path
from salt.ext import six

PROFILER_BINARY = '/usr/sbin/system_profiler'


def __virtual__():
    '''
    Check to see if the system_profiler binary is available
    '''
    PROFILER_BINARY = salt.utils.path.which('system_profiler')

    if PROFILER_BINARY:
        return True
    return (False, 'The system_profiler execution module cannot be loaded: '
            'system_profiler unavailable.')


def _call_system_profiler(datatype):
    '''
    Call out to system_profiler.  Return a dictionary
    of the stuff we are interested in.
    '''

    p = subprocess.Popen(
        [PROFILER_BINARY, '-detailLevel', 'full',
         '-xml', datatype], stdout=subprocess.PIPE)
    (sysprofresults, sysprof_stderr) = p.communicate(input=None)

    if six.PY2:
        plist = plistlib.readPlistFromString(sysprofresults)
    else:
        plist = plistlib.readPlistFromBytes(sysprofresults)

    try:
        apps = plist[0]['_items']
    except (IndexError, KeyError):
        apps = []

    return apps


def receipts():
    '''
    Return the results of a call to
    ``system_profiler -xml -detail full SPInstallHistoryDataType``
    as a dictionary.  Top-level keys of the dictionary
    are the names of each set of install receipts, since
    there can be multiple receipts with the same name.
    Contents of each key are a list of dictionaries.

    CLI Example:

    .. code-block:: bash

        salt '*' systemprofiler.receipts
    '''

    apps = _call_system_profiler('SPInstallHistoryDataType')

    appdict = {}

    for a in apps:
        details = dict(a)
        details.pop('_name')
        if 'install_date' in details:
            details['install_date'] = details['install_date'].strftime('%Y-%m-%d %H:%M:%S')
        if 'info' in details:
            try:
                details['info'] = '{0}: {1}'.format(details['info'][0],
                                                    details['info'][1].strftime('%Y-%m-%d %H:%M:%S'))
            except (IndexError, AttributeError):
                pass

        if a['_name'] not in appdict:
            appdict[a['_name']] = []

        appdict[a['_name']].append(details)

    return appdict


def applications():
    '''
    Return the results of a call to
    ``system_profiler -xml -detail full SPApplicationsDataType``
    as a dictionary.  Top-level keys of the dictionary
    are the names of each set of install receipts, since
    there can be multiple receipts with the same name.
    Contents of each key are a list of dictionaries.

    Note that this can take a long time depending on how many
    applications are installed on the target Mac.

    CLI Example:

    .. code-block:: bash

        salt '*' systemprofiler.applications
    '''

    apps = _call_system_profiler('SPApplicationsDataType')

    appdict = {}

    for a in apps:
        details = dict(a)
        details.pop('_name')
        if 'lastModified' in details:
            details['lastModified'] = details['lastModified'].strftime('%Y-%m-%d %H:%M:%S')
        if 'info' in details:
            try:
                details['info'] = '{0}: {1}'.format(details['info'][0],
                                                    details['info'][1].strftime('%Y-%m-%d %H:%M:%S'))
            except (IndexError, AttributeError):
                pass

        if a['_name'] not in appdict:
            appdict[a['_name']] = []

        appdict[a['_name']].append(details)

    return appdict



# -*- coding: utf-8 -*-
'''
A convenience system to manage reactors

Beginning in the 2017.7 release, the reactor runner requires that the reactor
system is running.  This is accomplished one of two ways, either
by having reactors configured or by including ``reactor`` in the
engine configuration for the Salt master.

    .. code-block:: yaml

    engines:
        - reactor

'''
# Import python libs
from __future__ import absolute_import, print_function, unicode_literals
import logging

# Import salt libs
import salt.utils.reactor
import salt.syspaths
import salt.utils.event
import salt.utils.process
from salt.ext.six import string_types

log = logging.getLogger(__name__)

__func_alias__ = {
    'list_': 'list',
}


def list_(saltenv='base', test=None):
    '''
    List currently configured reactors

    CLI Example:

    .. code-block:: bash

        salt-run reactor.list
    '''
    sevent = salt.utils.event.get_event(
            'master',
            __opts__['sock_dir'],
            __opts__['transport'],
            opts=__opts__,
            listen=True)

    __jid_event__.fire_event({}, 'salt/reactors/manage/list')

    results = sevent.get_event(wait=30, tag='salt/reactors/manage/list-results')
    reactors = results['reactors']
    return reactors


def add(event, reactors, saltenv='base', test=None):
    '''
    Add a new reactor

    CLI Example:

    .. code-block:: bash

        salt-run reactor.add 'salt/cloud/*/destroyed' reactors='/srv/reactor/destroy/*.sls'
    '''
    if isinstance(reactors, string_types):
        reactors = [reactors]

    sevent = salt.utils.event.get_event(
            'master',
            __opts__['sock_dir'],
            __opts__['transport'],
            opts=__opts__,
            listen=True)

    __jid_event__.fire_event({'event': event,
                              'reactors': reactors},
                             'salt/reactors/manage/add')

    res = sevent.get_event(wait=30, tag='salt/reactors/manage/add-complete')
    return res['result']


def delete(event, saltenv='base', test=None):
    '''
    Delete a reactor

    CLI Example:

    .. code-block:: bash

        salt-run reactor.delete 'salt/cloud/*/destroyed'
    '''
    sevent = salt.utils.event.get_event(
            'master',
            __opts__['sock_dir'],
            __opts__['transport'],
            opts=__opts__,
            listen=True)

    __jid_event__.fire_event({'event': event}, 'salt/reactors/manage/delete')

    res = sevent.get_event(wait=30, tag='salt/reactors/manage/delete-complete')
    return res['result']



# -*- coding: utf-8 -*-
'''
Module for Solaris 10's zoneadm

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:platform:      OmniOS,OpenIndiana,SmartOS,OpenSolaris,Solaris 10

.. versionadded:: 2017.7.0

.. warning::
    Oracle Solaris 11's zoneadm is not supported by this module!
'''
from __future__ import absolute_import, print_function, unicode_literals

# Import Python libs
import logging

# Import Salt libs
import salt.utils.path
import salt.utils.decorators
from salt.ext.six.moves import range

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = 'zoneadm'

# Function aliases
__func_alias__ = {
    'list_zones': 'list'
}


@salt.utils.decorators.memoize
def _is_globalzone():
    '''
    Check if we are running in the globalzone
    '''
    if not __grains__['kernel'] == 'SunOS':
        return False

    zonename = __salt__['cmd.run_all']('zonename')
    if zonename['retcode']:
        return False
    if zonename['stdout'] == 'global':
        return True

    return False


def _is_uuid(zone):
    '''
    Check if zone is actually a UUID
    '''
    return len(zone) == 36 and zone.index('-') == 8


def __virtual__():
    '''
    We are available if we are have zoneadm and are the global zone on
    Solaris 10, OmniOS, OpenIndiana, OpenSolaris, or Smartos.
    '''
    if _is_globalzone() and salt.utils.path.which('zoneadm'):
        if __grains__['os'] in ['OpenSolaris', 'SmartOS', 'OmniOS', 'OpenIndiana']:
            return __virtualname__
        elif __grains__['os'] == 'Oracle Solaris' and int(__grains__['osmajorrelease']) == 10:
            return __virtualname__
    return (
        False,
        '{0} module can only be loaded in a solaris globalzone.'.format(
            __virtualname__
        )
    )


def list_zones(verbose=True, installed=False, configured=False, hide_global=True):
    '''
    List all zones

    verbose : boolean
        display additional zone information
    installed : boolean
        include installed zones in output
    configured : boolean
        include configured zones in output
    hide_global : boolean
        do not include global zone

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.list
    '''
    zones = {}

    ## fetch zones
    header = 'zoneid:zonename:state:zonepath:uuid:brand:ip-type'.split(':')
    zone_data = __salt__['cmd.run_all']('zoneadm list -p -c')
    if zone_data['retcode'] == 0:
        for zone in zone_data['stdout'].splitlines():
            zone = zone.split(':')

            # create zone_t
            zone_t = {}
            for i in range(0, len(header)):
                zone_t[header[i]] = zone[i]

            # skip if global and hide_global
            if hide_global and zone_t['zonename'] == 'global':
                continue

            # skip installed and configured
            if not installed and zone_t['state'] == 'installed':
                continue
            if not configured and zone_t['state'] == 'configured':
                continue

            # update dict
            zones[zone_t['zonename']] = zone_t
            del zones[zone_t['zonename']]['zonename']

    return zones if verbose else sorted(zones.keys())


def boot(zone, single=False, altinit=None, smf_options=None):
    '''
    Boot (or activate) the specified zone.

    zone : string
        name or uuid of the zone
    single : boolean
        boots only to milestone svc:/milestone/single-user:default.
    altinit : string
        valid path to an alternative executable to be the primordial process.
    smf_options : string
        include two categories of options to control booting behavior of
        the service management facility: recovery options and messages options.

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.boot clementine
        salt '*' zoneadm.boot maeve single=True
        salt '*' zoneadm.boot teddy single=True smf_options=verbose
    '''
    ret = {'status': True}

    ## build boot_options
    boot_options = ''
    if single:
        boot_options = '-s {0}'.format(boot_options)
    if altinit:  # note: we cannot validate the path, as this is local to the zonepath.
        boot_options = '-i {0} {1}'.format(altinit, boot_options)
    if smf_options:
        boot_options = '-m {0} {1}'.format(smf_options, boot_options)
    if boot_options != '':
        boot_options = ' -- {0}'.format(boot_options.strip())

    ## execute boot
    res = __salt__['cmd.run_all']('zoneadm {zone} boot{boot_opts}'.format(
        zone='-u {0}'.format(zone) if _is_uuid(zone) else '-z {0}'.format(zone),
        boot_opts=boot_options,
    ))
    ret['status'] = res['retcode'] == 0
    ret['message'] = res['stdout'] if ret['status'] else res['stderr']
    ret['message'] = ret['message'].replace('zoneadm: ', '')
    if ret['message'] == '':
        del ret['message']

    return ret


def reboot(zone, single=False, altinit=None, smf_options=None):
    '''
    Restart the zone. This is equivalent to a halt boot sequence.

    zone : string
        name or uuid of the zone
    single : boolean
        boots only to milestone svc:/milestone/single-user:default.
    altinit : string
        valid path to an alternative executable to be the primordial process.
    smf_options : string
        include two categories of options to control booting behavior of
        the service management facility: recovery options and messages options.

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.reboot dolores
        salt '*' zoneadm.reboot teddy single=True
    '''
    ret = {'status': True}

    ## build boot_options
    boot_options = ''
    if single:
        boot_options = '-s {0}'.format(boot_options)
    if altinit:  # note: we cannot validate the path, as this is local to the zonepath.
        boot_options = '-i {0} {1}'.format(altinit, boot_options)
    if smf_options:
        boot_options = '-m {0} {1}'.format(smf_options, boot_options)
    if boot_options != '':
        boot_options = ' -- {0}'.format(boot_options.strip())

    ## execute reboot
    res = __salt__['cmd.run_all']('zoneadm {zone} reboot{boot_opts}'.format(
        zone='-u {0}'.format(zone) if _is_uuid(zone) else '-z {0}'.format(zone),
        boot_opts=boot_options,
    ))
    ret['status'] = res['retcode'] == 0
    ret['message'] = res['stdout'] if ret['status'] else res['stderr']
    ret['message'] = ret['message'].replace('zoneadm: ', '')
    if ret['message'] == '':
        del ret['message']

    return ret


def halt(zone):
    '''
    Halt the specified zone.

    zone : string
        name or uuid of the zone

    .. note::
        To cleanly shutdown the zone use the shutdown function.

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.halt hector
    '''
    ret = {'status': True}

    ## halt zone
    res = __salt__['cmd.run_all']('zoneadm {zone} halt'.format(
        zone='-u {0}'.format(zone) if _is_uuid(zone) else '-z {0}'.format(zone),
    ))
    ret['status'] = res['retcode'] == 0
    ret['message'] = res['stdout'] if ret['status'] else res['stderr']
    ret['message'] = ret['message'].replace('zoneadm: ', '')
    if ret['message'] == '':
        del ret['message']

    return ret


def shutdown(zone, reboot=False, single=False, altinit=None, smf_options=None):
    '''
    Gracefully shutdown the specified zone.

    zone : string
        name or uuid of the zone
    reboot : boolean
        reboot zone after shutdown (equivalent of shutdown -i6 -g0 -y)
    single : boolean
        boots only to milestone svc:/milestone/single-user:default.
    altinit : string
        valid path to an alternative executable to be the primordial process.
    smf_options : string
        include two categories of options to control booting behavior of
        the service management facility: recovery options and messages options.

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.shutdown peter
        salt '*' zoneadm.shutdown armistice reboot=True
    '''
    ret = {'status': True}

    ## build boot_options
    boot_options = ''
    if single:
        boot_options = '-s {0}'.format(boot_options)
    if altinit:  # note: we cannot validate the path, as this is local to the zonepath.
        boot_options = '-i {0} {1}'.format(altinit, boot_options)
    if smf_options:
        boot_options = '-m {0} {1}'.format(smf_options, boot_options)
    if boot_options != '':
        boot_options = ' -- {0}'.format(boot_options.strip())

    ## shutdown zone
    res = __salt__['cmd.run_all']('zoneadm {zone} shutdown{reboot}{boot_opts}'.format(
        zone='-u {0}'.format(zone) if _is_uuid(zone) else '-z {0}'.format(zone),
        reboot=' -r' if reboot else '',
        boot_opts=boot_options,
    ))
    ret['status'] = res['retcode'] == 0
    ret['message'] = res['stdout'] if ret['status'] else res['stderr']
    ret['message'] = ret['message'].replace('zoneadm: ', '')
    if ret['message'] == '':
        del ret['message']

    return ret


def detach(zone):
    '''
    Detach the specified zone.

    zone : string
        name or uuid of the zone

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.detach kissy
    '''
    ret = {'status': True}

    ## detach zone
    res = __salt__['cmd.run_all']('zoneadm {zone} detach'.format(
        zone='-u {0}'.format(zone) if _is_uuid(zone) else '-z {0}'.format(zone),
    ))
    ret['status'] = res['retcode'] == 0
    ret['message'] = res['stdout'] if ret['status'] else res['stderr']
    ret['message'] = ret['message'].replace('zoneadm: ', '')
    if ret['message'] == '':
        del ret['message']

    return ret


def attach(zone, force=False, brand_opts=None):
    '''
    Attach the specified zone.

    zone : string
        name of the zone
    force : boolean
        force the zone into the "installed" state with no validation
    brand_opts : string
        brand specific options to pass

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.attach lawrence
        salt '*' zoneadm.attach lawrence True
    '''
    ret = {'status': True}

    ## attach zone
    res = __salt__['cmd.run_all']('zoneadm -z {zone} attach{force}{brand_opts}'.format(
        zone=zone,
        force=' -F' if force else '',
        brand_opts=' {0}'.format(brand_opts) if brand_opts else '',
    ))
    ret['status'] = res['retcode'] == 0
    ret['message'] = res['stdout'] if ret['status'] else res['stderr']
    ret['message'] = ret['message'].replace('zoneadm: ', '')
    if ret['message'] == '':
        del ret['message']

    return ret


def ready(zone):
    '''
    Prepares a zone for running applications.

    zone : string
        name or uuid of the zone

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.ready clementine
    '''
    ret = {'status': True}

    ## ready zone
    res = __salt__['cmd.run_all']('zoneadm {zone} ready'.format(
        zone='-u {0}'.format(zone) if _is_uuid(zone) else '-z {0}'.format(zone),
    ))
    ret['status'] = res['retcode'] == 0
    ret['message'] = res['stdout'] if ret['status'] else res['stderr']
    ret['message'] = ret['message'].replace('zoneadm: ', '')
    if ret['message'] == '':
        del ret['message']

    return ret


def verify(zone):
    '''
    Check to make sure the configuration of the specified
    zone can safely be installed on the machine.

    zone : string
        name of the zone

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.verify dolores
    '''
    ret = {'status': True}

    ## verify zone
    res = __salt__['cmd.run_all']('zoneadm -z {zone} verify'.format(
        zone=zone,
    ))
    ret['status'] = res['retcode'] == 0
    ret['message'] = res['stdout'] if ret['status'] else res['stderr']
    ret['message'] = ret['message'].replace('zoneadm: ', '')
    if ret['message'] == '':
        del ret['message']

    return ret


def move(zone, zonepath):
    '''
    Move zone to new zonepath.

    zone : string
        name or uuid of the zone
    zonepath : string
        new zonepath

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.move meave /sweetwater/meave
    '''
    ret = {'status': True}

    ## verify zone
    res = __salt__['cmd.run_all']('zoneadm {zone} move {path}'.format(
        zone='-u {0}'.format(zone) if _is_uuid(zone) else '-z {0}'.format(zone),
        path=zonepath,
    ))
    ret['status'] = res['retcode'] == 0
    ret['message'] = res['stdout'] if ret['status'] else res['stderr']
    ret['message'] = ret['message'].replace('zoneadm: ', '')
    if ret['message'] == '':
        del ret['message']

    return ret


def uninstall(zone):
    '''
    Uninstall the specified zone from the system.

    zone : string
        name or uuid of the zone

    .. warning::
        The -F flag is always used to avoid the prompts when uninstalling.

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.uninstall teddy
    '''
    ret = {'status': True}

    ## uninstall zone
    res = __salt__['cmd.run_all']('zoneadm {zone} uninstall -F'.format(
        zone='-u {0}'.format(zone) if _is_uuid(zone) else '-z {0}'.format(zone),
    ))
    ret['status'] = res['retcode'] == 0
    ret['message'] = res['stdout'] if ret['status'] else res['stderr']
    ret['message'] = ret['message'].replace('zoneadm: ', '')
    if ret['message'] == '':
        del ret['message']

    return ret


def install(zone, nodataset=False, brand_opts=None):
    '''
    Install the specified zone from the system.

    zone : string
        name of the zone
    nodataset : boolean
        do not create a ZFS file system
    brand_opts : string
        brand specific options to pass

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.install dolores
        salt '*' zoneadm.install teddy True
    '''
    ret = {'status': True}

    ## install zone
    res = __salt__['cmd.run_all']('zoneadm -z {zone} install{nodataset}{brand_opts}'.format(
        zone=zone,
        nodataset=' -x nodataset' if nodataset else '',
        brand_opts=' {0}'.format(brand_opts) if brand_opts else '',
    ))
    ret['status'] = res['retcode'] == 0
    ret['message'] = res['stdout'] if ret['status'] else res['stderr']
    ret['message'] = ret['message'].replace('zoneadm: ', '')
    if ret['message'] == '':
        del ret['message']

    return ret


def clone(zone, source, snapshot=None):
    '''
    Install a zone by copying an existing installed zone.

    zone : string
        name of the zone
    source : string
        zone to clone from
    snapshot : string
        optional name of snapshot to use as source

    CLI Example:

    .. code-block:: bash

        salt '*' zoneadm.clone clementine dolores
    '''
    ret = {'status': True}

    ## install zone
    res = __salt__['cmd.run_all']('zoneadm -z {zone} clone {snapshot}{source}'.format(
        zone=zone,
        source=source,
        snapshot='-s {0} '.format(snapshot) if snapshot else '',
    ))
    ret['status'] = res['retcode'] == 0
    ret['message'] = res['stdout'] if ret['status'] else res['stderr']
    ret['message'] = ret['message'].replace('zoneadm: ', '')
    if ret['message'] == '':
        del ret['message']

    return ret

# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



# -*- coding: utf-8 -*-
'''
Return data to a PostgreSQL server with json data stored in Pg's jsonb data type

:maintainer:    Dave Boucha <dave@saltstack.com>, Seth House <shouse@saltstack.com>, C. R. Oldham <cr@saltstack.com>
:maturity:      Stable
:depends:       python-psycopg2
:platform:      all

.. note::
    There are three PostgreSQL returners.  Any can function as an external
    :ref:`master job cache <external-job-cache>`. but each has different
    features.  SaltStack recommends
    :mod:`returners.pgjsonb <salt.returners.pgjsonb>` if you are working with
    a version of PostgreSQL that has the appropriate native binary JSON types.
    Otherwise, review
    :mod:`returners.postgres <salt.returners.postgres>` and
    :mod:`returners.postgres_local_cache <salt.returners.postgres_local_cache>`
    to see which module best suits your particular needs.

To enable this returner, the minion will need the python client for PostgreSQL
installed and the following values configured in the minion or master
config. These are the defaults:

.. code-block:: yaml

    returner.pgjsonb.host: 'salt'
    returner.pgjsonb.user: 'salt'
    returner.pgjsonb.pass: 'salt'
    returner.pgjsonb.db: 'salt'
    returner.pgjsonb.port: 5432

SSL is optional. The defaults are set to None. If you do not want to use SSL,
either exclude these options or set them to None.

.. code-block:: yaml

    returner.pgjsonb.sslmode: None
    returner.pgjsonb.sslcert: None
    returner.pgjsonb.sslkey: None
    returner.pgjsonb.sslrootcert: None
    returner.pgjsonb.sslcrl: None

.. versionadded:: 2017.5.0

Alternative configuration values can be used by prefacing the configuration
with `alternative.`. Any values not found in the alternative configuration will
be pulled from the default location. As stated above, SSL configuration is
optional. The following ssl options are simply for illustration purposes:

.. code-block:: yaml

    alternative.pgjsonb.host: 'salt'
    alternative.pgjsonb.user: 'salt'
    alternative.pgjsonb.pass: 'salt'
    alternative.pgjsonb.db: 'salt'
    alternative.pgjsonb.port: 5432
    alternative.pgjsonb.ssl_ca: '/etc/pki/mysql/certs/localhost.pem'
    alternative.pgjsonb.ssl_cert: '/etc/pki/mysql/certs/localhost.crt'
    alternative.pgjsonb.ssl_key: '/etc/pki/mysql/certs/localhost.key'

Use the following Pg database schema:

.. code-block:: sql

    CREATE DATABASE  salt
      WITH ENCODING 'utf-8';

    --
    -- Table structure for table `jids`
    --
    DROP TABLE IF EXISTS jids;
    CREATE TABLE jids (
       jid varchar(255) NOT NULL primary key,
       load jsonb NOT NULL
    );
    CREATE INDEX idx_jids_jsonb on jids
           USING gin (load)
           WITH (fastupdate=on);

    --
    -- Table structure for table `salt_returns`
    --

    DROP TABLE IF EXISTS salt_returns;
    CREATE TABLE salt_returns (
      fun varchar(50) NOT NULL,
      jid varchar(255) NOT NULL,
      return jsonb NOT NULL,
      id varchar(255) NOT NULL,
      success varchar(10) NOT NULL,
      full_ret jsonb NOT NULL,
      alter_time TIMESTAMP WITH TIME ZONE DEFAULT NOW());

    CREATE INDEX idx_salt_returns_id ON salt_returns (id);
    CREATE INDEX idx_salt_returns_jid ON salt_returns (jid);
    CREATE INDEX idx_salt_returns_fun ON salt_returns (fun);
    CREATE INDEX idx_salt_returns_return ON salt_returns
        USING gin (return) with (fastupdate=on);
    CREATE INDEX idx_salt_returns_full_ret ON salt_returns
        USING gin (full_ret) with (fastupdate=on);

    --
    -- Table structure for table `salt_events`
    --

    DROP TABLE IF EXISTS salt_events;
    DROP SEQUENCE IF EXISTS seq_salt_events_id;
    CREATE SEQUENCE seq_salt_events_id;
    CREATE TABLE salt_events (
        id BIGINT NOT NULL UNIQUE DEFAULT nextval('seq_salt_events_id'),
        tag varchar(255) NOT NULL,
        data jsonb NOT NULL,
        alter_time TIMESTAMP WITH TIME ZONE DEFAULT NOW(),
        master_id varchar(255) NOT NULL);

    CREATE INDEX idx_salt_events_tag on
        salt_events (tag);
    CREATE INDEX idx_salt_events_data ON salt_events
        USING gin (data) with (fastupdate=on);

Required python modules: Psycopg2

To use this returner, append '--return pgjsonb' to the salt command.

.. code-block:: bash

    salt '*' test.ping --return pgjsonb

To use the alternative configuration, append '--return_config alternative' to the salt command.

.. versionadded:: 2015.5.0

.. code-block:: bash

    salt '*' test.ping --return pgjsonb --return_config alternative

To override individual configuration items, append --return_kwargs '{"key:": "value"}' to the salt command.

.. versionadded:: 2016.3.0

.. code-block:: bash

    salt '*' test.ping --return pgjsonb --return_kwargs '{"db": "another-salt"}'

'''
from __future__ import absolute_import, print_function, unicode_literals
# Let's not allow PyLint complain about string substitution
# pylint: disable=W1321,E1321

# Import python libs
from contextlib import contextmanager
import sys
import time
import logging

# Import salt libs
import salt.returners
import salt.utils.jid
import salt.exceptions
from salt.ext import six

# Import third party libs
try:
    import psycopg2
    import psycopg2.extras
    HAS_PG = True
except ImportError:
    HAS_PG = False

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = 'pgjsonb'


def __virtual__():
    if not HAS_PG:
        return False, 'Could not import pgjsonb returner; python-psycopg2 is not installed.'
    return True


def _get_options(ret=None):
    '''
    Returns options used for the MySQL connection.
    '''
    defaults = {
        'host': 'localhost',
        'user': 'salt',
        'pass': 'salt',
        'db': 'salt',
        'port': 5432
    }

    attrs = {
        'host': 'host',
        'user': 'user',
        'pass': 'pass',
        'db': 'db',
        'port': 'port',
        'sslmode': 'sslmode',
        'sslcert': 'sslcert',
        'sslkey': 'sslkey',
        'sslrootcert': 'sslrootcert',
        'sslcrl': 'sslcrl',
    }

    _options = salt.returners.get_returner_options('returner.{0}'.format(__virtualname__),
                                                   ret,
                                                   attrs,
                                                   __salt__=__salt__,
                                                   __opts__=__opts__,
                                                   defaults=defaults)
    # Ensure port is an int
    if 'port' in _options:
        _options['port'] = int(_options['port'])
    return _options


@contextmanager
def _get_serv(ret=None, commit=False):
    '''
    Return a Pg cursor
    '''
    _options = _get_options(ret)
    try:
        # An empty ssl_options dictionary passed to MySQLdb.connect will
        # effectively connect w/o SSL.
        ssl_options = {
            k: v for k, v in six.iteritems(_options)
            if k in ['sslmode', 'sslcert', 'sslkey', 'sslrootcert', 'sslcrl']
        }
        conn = psycopg2.connect(
            host=_options.get('host'),
            port=_options.get('port'),
            dbname=_options.get('db'),
            user=_options.get('user'),
            password=_options.get('pass'),
            **ssl_options
        )
    except psycopg2.OperationalError as exc:
        raise salt.exceptions.SaltMasterError('pgjsonb returner could not connect to database: {exc}'.format(exc=exc))

    cursor = conn.cursor()

    try:
        yield cursor
    except psycopg2.DatabaseError as err:
        error = err.args
        sys.stderr.write(six.text_type(error))
        cursor.execute("ROLLBACK")
        raise err
    else:
        if commit:
            cursor.execute("COMMIT")
        else:
            cursor.execute("ROLLBACK")
    finally:
        conn.close()


def returner(ret):
    '''
    Return data to a Pg server
    '''
    try:
        with _get_serv(ret, commit=True) as cur:
            sql = '''INSERT INTO salt_returns
                    (fun, jid, return, id, success, full_ret, alter_time)
                    VALUES (%s, %s, %s, %s, %s, %s, to_timestamp(%s))'''

            cur.execute(sql, (ret['fun'], ret['jid'],
                              psycopg2.extras.Json(ret['return']),
                              ret['id'],
                              ret.get('success', False),
                              psycopg2.extras.Json(ret),
                              time.time()))
    except salt.exceptions.SaltMasterError:
        log.critical('Could not store return with pgjsonb returner. PostgreSQL server unavailable.')


def event_return(events):
    '''
    Return event to Pg server

    Requires that configuration be enabled via 'event_return'
    option in master config.
    '''
    with _get_serv(events, commit=True) as cur:
        for event in events:
            tag = event.get('tag', '')
            data = event.get('data', '')
            sql = '''INSERT INTO salt_events (tag, data, master_id, alter_time)
                     VALUES (%s, %s, %s, to_timestamp(%s))'''
            cur.execute(sql, (tag, psycopg2.extras.Json(data),
                              __opts__['id'], time.time()))


def save_load(jid, load, minions=None):
    '''
    Save the load to the specified jid id
    '''
    with _get_serv(commit=True) as cur:

        sql = '''INSERT INTO jids
               (jid, load)
                VALUES (%s, %s)'''

        try:
            cur.execute(sql, (jid, psycopg2.extras.Json(load)))
        except psycopg2.IntegrityError:
            # https://github.com/saltstack/salt/issues/22171
            # Without this try:except: we get tons of duplicate entry errors
            # which result in job returns not being stored properly
            pass


def save_minions(jid, minions, syndic_id=None):  # pylint: disable=unused-argument
    '''
    Included for API consistency
    '''
    pass


def get_load(jid):
    '''
    Return the load data that marks a specified jid
    '''
    with _get_serv(ret=None, commit=True) as cur:

        sql = '''SELECT load FROM jids WHERE jid = %s;'''
        cur.execute(sql, (jid,))
        data = cur.fetchone()
        if data:
            return data[0]
        return {}


def get_jid(jid):
    '''
    Return the information returned when the specified job id was executed
    '''
    with _get_serv(ret=None, commit=True) as cur:

        sql = '''SELECT id, full_ret FROM salt_returns
                WHERE jid = %s'''

        cur.execute(sql, (jid,))
        data = cur.fetchall()
        ret = {}
        if data:
            for minion, full_ret in data:
                ret[minion] = full_ret
        return ret


def get_fun(fun):
    '''
    Return a dict of the last function called for all minions
    '''
    with _get_serv(ret=None, commit=True) as cur:

        sql = '''SELECT s.id,s.jid, s.full_ret
                FROM salt_returns s
                JOIN ( SELECT MAX(`jid`) as jid
                    from salt_returns GROUP BY fun, id) max
                ON s.jid = max.jid
                WHERE s.fun = %s
                '''

        cur.execute(sql, (fun,))
        data = cur.fetchall()

        ret = {}
        if data:
            for minion, _, full_ret in data:
                ret[minion] = full_ret
        return ret


def get_jids():
    '''
    Return a list of all job ids
    '''
    with _get_serv(ret=None, commit=True) as cur:

        sql = '''SELECT jid, load
                FROM jids'''

        cur.execute(sql)
        data = cur.fetchall()
        ret = {}
        for jid, load in data:
            ret[jid] = salt.utils.jid.format_jid_instance(jid, load)
        return ret


def get_minions():
    '''
    Return a list of minions
    '''
    with _get_serv(ret=None, commit=True) as cur:

        sql = '''SELECT DISTINCT id
                FROM salt_returns'''

        cur.execute(sql)
        data = cur.fetchall()
        ret = []
        for minion in data:
            ret.append(minion[0])
        return ret


def prep_jid(nocache=False, passed_jid=None):  # pylint: disable=unused-argument
    '''
    Do any work necessary to prepare a JID, including sending a custom id
    '''
    return passed_jid if passed_jid is not None else salt.utils.jid.gen_jid(__opts__)



# -*- coding: utf-8 -*-
'''
Stateconf System
================

The stateconf system is intended for use only with the stateconf renderer. This
State module presents the set function. This function does not execute any
functionality, but is used to interact with the stateconf renderer.
'''
from __future__ import absolute_import, print_function, unicode_literals


def _no_op(name, **kwargs):
    '''
    No-op state to support state config via the stateconf renderer.
    '''
    return dict(name=name, result=True, changes={}, comment='')

set = context = _no_op  # pylint: disable=C0103



# -*- coding: utf-8 -*-
'''
Contains systemd related help files
'''
# import python libs
from __future__ import absolute_import, print_function, unicode_literals
import logging
import os
import subprocess

# Import Salt libs
from salt.exceptions import SaltInvocationError
import salt.utils.stringutils

log = logging.getLogger(__name__)


def booted(context=None):
    '''
    Return True if the system was booted with systemd, False otherwise.  If the
    loader context dict ``__context__`` is passed, this function will set the
    ``salt.utils.systemd.booted`` key to represent if systemd is running and
    keep the logic below from needing to be run again during the same salt run.
    '''
    contextkey = 'salt.utils.systemd.booted'
    if isinstance(context, dict):
        # Can't put this if block on the same line as the above if block,
        # because it willl break the elif below.
        if contextkey in context:
            return context[contextkey]
    elif context is not None:
        raise SaltInvocationError('context must be a dictionary if passed')

    try:
        # This check does the same as sd_booted() from libsystemd-daemon:
        # http://www.freedesktop.org/software/systemd/man/sd_booted.html
        ret = bool(os.stat('/run/systemd/system'))
    except OSError:
        ret = False

    try:
        context[contextkey] = ret
    except TypeError:
        pass

    return ret


def version(context=None):
    '''
    Attempts to run systemctl --version. Returns None if unable to determine
    version.
    '''
    contextkey = 'salt.utils.systemd.version'
    if isinstance(context, dict):
        # Can't put this if block on the same line as the above if block,
        # because it will break the elif below.
        if contextkey in context:
            return context[contextkey]
    elif context is not None:
        raise SaltInvocationError('context must be a dictionary if passed')
    stdout = subprocess.Popen(
        ['systemctl', '--version'],
        close_fds=True,
        stdout=subprocess.PIPE, stderr=subprocess.STDOUT).communicate()[0]
    outstr = salt.utils.stringutils.to_str(stdout)
    try:
        ret = int(outstr.splitlines()[0].split()[-1])
    except (IndexError, ValueError):
        log.error(
            'Unable to determine systemd version from systemctl '
            '--version, output follows:\n%s', outstr
        )
        return None
    else:
        try:
            context[contextkey] = ret
        except TypeError:
            pass
        return ret


def has_scope(context=None):
    '''
    Scopes were introduced in systemd 205, this function returns a boolean
    which is true when the minion is systemd-booted and running systemd>=205.
    '''
    if not booted(context):
        return False
    _sd_version = version(context)
    if _sd_version is None:
        return False
    return _sd_version >= 205



# -*- coding: utf-8 -*-
'''
Modify, retrieve, or delete values from OpenStack configuration files.

:maintainer: Jeffrey C. Ollie <jeff@ocjtech.us>
:maturity: new
:depends:
:platform: linux

'''
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals
import shlex

try:
    import pipes
    HAS_DEPS = True
except ImportError:
    HAS_DEPS = False

if hasattr(shlex, 'quote'):
    _quote = shlex.quote
elif HAS_DEPS and hasattr(pipes, 'quote'):
    _quote = pipes.quote
else:
    _quote = None

# Import Salt libs
from salt.ext import six
import salt.utils.decorators.path
import salt.exceptions

# Don't shadow built-in's.
__func_alias__ = {
    'set_': 'set'
}


def __virtual__():
    if _quote is None and not HAS_DEPS:
        return False
    return True


def _fallback(*args, **kw):
    return 'The "openstack-config" command needs to be installed for this function to work.  Typically this is included in the "openstack-utils" package.'


@salt.utils.decorators.path.which('openstack-config')
def set_(filename, section, parameter, value):
    '''
    Set a value in an OpenStack configuration file.

    filename
        The full path to the configuration file

    section
        The section in which the parameter will be set

    parameter
        The parameter to change

    value
        The value to set

    CLI Example:

    .. code-block:: bash

        salt-call openstack_config.set /etc/keystone/keystone.conf sql connection foo
    '''

    filename = _quote(filename)
    section = _quote(section)
    parameter = _quote(parameter)
    value = _quote(six.text_type(value))

    result = __salt__['cmd.run_all'](
            'openstack-config --set {0} {1} {2} {3}'.format(
                filename, section, parameter, value
                ),
            python_shell=False,
            )

    if result['retcode'] == 0:
        return result['stdout']
    else:
        raise salt.exceptions.CommandExecutionError(result['stderr'])


@salt.utils.decorators.path.which('openstack-config')
def get(filename, section, parameter):
    '''
    Get a value from an OpenStack configuration file.

    filename
        The full path to the configuration file

    section
        The section from which to search for the parameter

    parameter
        The parameter to return

    CLI Example:

    .. code-block:: bash

        salt-call openstack_config.get /etc/keystone/keystone.conf sql connection

    '''

    filename = _quote(filename)
    section = _quote(section)
    parameter = _quote(parameter)

    result = __salt__['cmd.run_all'](
            'openstack-config --get {0} {1} {2}'.format(
                filename, section, parameter
                ),
            python_shell=False,
            )

    if result['retcode'] == 0:
        return result['stdout']
    else:
        raise salt.exceptions.CommandExecutionError(result['stderr'])


@salt.utils.decorators.path.which('openstack-config')
def delete(filename, section, parameter):
    '''
    Delete a value from an OpenStack configuration file.

    filename
        The full path to the configuration file

    section
        The section from which to delete the parameter

    parameter
        The parameter to delete

    CLI Example:

    .. code-block:: bash

        salt-call openstack_config.delete /etc/keystone/keystone.conf sql connection
    '''

    filename = _quote(filename)
    section = _quote(section)
    parameter = _quote(parameter)

    result = __salt__['cmd.run_all'](
            'openstack-config --del {0} {1} {2}'.format(
                filename, section, parameter
                ),
            python_shell=False,
            )

    if result['retcode'] == 0:
        return result['stdout']
    else:
        raise salt.exceptions.CommandExecutionError(result['stderr'])



# -*- coding: utf-8 -*-
'''
Installing of mac pkg files
===========================

Install any kind of pkg, dmg or app file on macOS:

.. code-block:: yaml

    /mnt/test.pkg:
      macpackage.installed:
        - store: True

    /mnt/test.dmg:
      macpackage.installed:
        - dmg: True

    /mnt/xcode.dmg:
      macpackage.installed:
        - dmg: True
        - app: True
        - target: /Applications/Xcode.app
        - version_check: xcodebuild -version=Xcode 7.1\\n.*7B91b

'''
# Import Python libs
from __future__ import absolute_import, unicode_literals, print_function
import logging
import os
import re

# Import Salt libs
import salt.utils.platform
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)
__virtualname__ = "macpackage"


def __virtual__():
    '''
    Only work on Mac OS
    '''
    if salt.utils.platform.is_darwin():
        return __virtualname__
    return False


def installed(name, target="LocalSystem", dmg=False, store=False, app=False, mpkg=False, user=None, onlyif=None,
              unless=None, force=False, allow_untrusted=False, version_check=None):
    '''
    Install a Mac OS Package from a pkg or dmg file, if given a dmg file it
    will first be mounted in a temporary location

    name
        The pkg or dmg file to install

    target
        The location in which to install the package. This can be a path or LocalSystem

    dmg
        Is the given file a dmg file?

    store
        Should the pkg be installed as if it was from the Mac OS Store?

    app
        Is the file a .app? If so then we'll just copy that to /Applications/ or the given
        target

    mpkg
        Is the file a .mpkg? If so then we'll check all of the .pkg files found are installed

    user
        Name of the user performing the unless or onlyif checks

    onlyif
        A command to run as a check, run the named command only if the command
        passed to the ``onlyif`` option returns true

    unless
        A command to run as a check, only run the named command if the command
        passed to the ``unless`` option returns false

    force
        Force the package to be installed even if its already been found installed

    allow_untrusted
        Allow the installation of untrusted packages

    version_check
        The command and version that we want to check against, the version number can use regex.

        .. code-block:: yaml

            version_check: python --version_check=2.7.[0-9]

    '''
    ret = {'name': name,
           'result': True,
           'comment': '',
           'changes': {}}
    found = []
    installing = []

    real_pkg = name

    # Check onlyif, unless first
    run_check_cmd_kwargs = {'runas': user, 'python_shell': True}
    if 'shell' in __grains__:
        run_check_cmd_kwargs['shell'] = __grains__['shell']

    cret = _mod_run_check(run_check_cmd_kwargs, onlyif, unless)

    if isinstance(cret, dict):
        ret.update(cret)
        return ret

    # Check version info
    if version_check is not None:
        split = version_check.split("=")
        if len(split) == 2:
            version_cmd = split[0]
            expected_version = split[1]
            try:
                version_out = __salt__['cmd.run'](version_cmd, output_loglevel="quiet", ignore_retcode=True)
                version_out = version_out.strip()
            except CommandExecutionError:
                version_out = ""

            if re.match(expected_version, version_out) is not None:
                ret['comment'] += "Version already matches {0}".format(expected_version)
                return ret
            else:
                ret['comment'] += "Version {0} doesn't match {1}. ".format(version_out, expected_version)

    if app and target == "LocalSystem":
        target = "/Applications/"

    # Mount the dmg first
    mount_point = None
    if dmg:
        out, mount_point = __salt__['macpackage.mount'](name)
        if 'attach failed' in out:
            ret['result'] = False
            ret['comment'] += 'Unable to mount {0}'.format(name)
            return ret

        if app:
            real_pkg = mount_point + "/*.app"
        elif mpkg:
            real_pkg = mount_point + "/*.mpkg"
        else:
            real_pkg = mount_point + "/*.pkg"

    try:
        # Check if we have already installed this
        if app:
            if dmg:
                # Run with python shell due to the wildcard
                cmd = 'ls -d *.app'
                out = __salt__['cmd.run'](cmd, cwd=mount_point, python_shell=True)

                if '.app' not in out:
                    ret['result'] = False
                    ret['comment'] += 'Unable to find .app in {0}'.format(mount_point)
                    return ret
                else:
                    pkg_ids = out.split("\n")
            else:
                pkg_ids = [os.path.basename(name)]
                mount_point = os.path.dirname(name)

            if onlyif is None and unless is None and version_check is None:
                for p in pkg_ids:
                    if target[-4:] == ".app":
                        install_dir = target
                    else:
                        install_dir = os.path.join(target, p)
                    if os.path.exists(install_dir) and force is False:
                        found.append(p)
                    else:
                        installing.append(p)
            else:
                installing = pkg_ids
        else:
            installed_pkgs = __salt__['macpackage.installed_pkgs']()

            if mpkg:
                pkg_ids = __salt__['macpackage.get_mpkg_ids'](real_pkg)
            else:
                pkg_ids = __salt__['macpackage.get_pkg_id'](real_pkg)

            if len(pkg_ids) > 0:
                for p in pkg_ids:
                    if p in installed_pkgs and force is False:
                        found.append(p)
                    else:
                        installing.append(p)
                if len(pkg_ids) == len(found):
                    return ret

        if app:
            def failed_pkg(f_pkg):
                ret['result'] = False
                ret['comment'] += '{0} failed to install: {1}'.format(name, out)

                if 'failed' in ret['changes']:
                    ret['changes']['failed'].append(f_pkg)
                else:
                    ret['changes']['failed'] = [f_pkg]

            for app in installing:
                try:
                    log.info('Copying {0} to {1}'.format(app, target))

                    out = __salt__['macpackage.install_app'](os.path.join(mount_point, app), target)

                    if len(out) != 0:
                        failed_pkg(app)
                    else:
                        ret['comment'] += '{0} installed'.format(app)
                        if 'installed' in ret['changes']:
                            ret['changes']['installed'].append(app)
                        else:
                            ret['changes']['installed'] = [app]

                except OSError:
                    failed_pkg(app)
        else:
            out = __salt__['macpackage.install'](real_pkg, target, store, allow_untrusted)

            if out['retcode'] != 0:
                ret['result'] = False
                ret['comment'] += '. {0} failed to install: {1}'.format(name, out)
            else:
                ret['comment'] += '{0} installed'.format(name)
                ret['changes']['installed'] = installing

    finally:
        if dmg:
            # Unmount to be kind
            __salt__['macpackage.unmount'](mount_point)

    return ret


def _mod_run_check(cmd_kwargs, onlyif, unless):
    '''
    Execute the onlyif and unless logic.
    Return a result dict if:
    * onlyif failed (onlyif != 0)
    * unless succeeded (unless == 0)
    else return True
    '''
    if onlyif:
        if __salt__['cmd.retcode'](onlyif, **cmd_kwargs) != 0:
            return {'comment': 'onlyif condition is false',
                    'skip_watch': True,
                    'result': True}

    if unless:
        if __salt__['cmd.retcode'](unless, **cmd_kwargs) == 0:
            return {'comment': 'unless condition is true',
                    'skip_watch': True,
                    'result': True}

    # No reason to stop, return True
    return True



# -*- coding: utf-8 -*-
'''
Manage DynamoDB Tables
======================

.. versionadded:: 2015.5.0

Create and destroy DynamoDB tables. Be aware that this interacts with Amazon's
services, and so may incur charges.

This module uses ``boto``, which can be installed via package, or pip.

This module accepts explicit DynamoDB credentials but can also utilize
IAM roles assigned to the instance through Instance Profiles. Dynamic
credentials are then automatically obtained from AWS API and no further
configuration is necessary. More information available `here
<http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html>`_.

If IAM roles are not used you need to specify them either in a pillar file or
in the minion's config file:

.. code-block:: yaml

    keyid: GKTADJGHEIQSXMKKRBJ08H
    key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
    region: us-east-1

It's also possible to specify ``key``, ``keyid`` and ``region`` via a
profile, either passed in as a dict, or as a string to pull from
pillars or minion config:

.. code-block:: yaml

    myprofile:
        keyid: GKTADJGHEIQSXMKKRBJ08H
        key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

.. code-block:: yaml

    Ensure DynamoDB table does not exist:
      boto_dynamodb.absent:
        - table_name: new_table
        - keyid: GKTADJGHEIQSXMKKRBJ08H
        - key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
        - region: us-east-1

    Ensure DynamoDB table exists:
      boto_dynamodb.present:
        - table_name: new_table
        - read_capacity_units: 1
        - write_capacity_units: 2
        - hash_key: primary_id
        - hash_key_data_type: N
        - range_key: start_timestamp
        - range_key_data_type: N
        - keyid: GKTADJGHEIQSXMKKRBJ08H
        - key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
        - region: us-east-1
        - local_indexes:
            - index:
                - name: "primary_id_end_timestamp_index"
                - hash_key: primary_id
                - hash_key_data_type: N
                - range_key: end_timestamp
                - range_key_data_type: N
        - global_indexes:
            - index:
                - name: "name_end_timestamp_index"
                - hash_key: name
                - hash_key_data_type: S
                - range_key: end_timestamp
                - range_key_data_type: N
                - read_capacity_units: 3
                - write_capacity_units: 4

It's possible to specify cloudwatch alarms that will be setup along with the
DynamoDB table. Note the alarm name will be defined by the name attribute
provided, plus the DynamoDB resource name.

.. code-block:: yaml

    Ensure DynamoDB table exists:
      boto_dynamodb.present:
        - name: new_table
        - read_capacity_units: 1
        - write_capacity_units: 2
        - hash_key: primary_id
        - hash_key_data_type: N
        - range_key: start_timestamp
        - range_key_data_type: N
        - alarms:
             ConsumedWriteCapacityUnits:
                name: 'DynamoDB ConsumedWriteCapacityUnits **MANAGED BY SALT**'
                attributes:
                  metric: ConsumedWriteCapacityUnits
                  namespace: AWS/DynamoDB
                  statistic: Sum
                  comparison: '>='
                  # threshold_percent is used to calculate the actual threshold
                  # based on the provisioned capacity for the table.
                  threshold_percent: 0.75
                  period: 300
                  evaluation_periods: 2
                  unit: Count
                  description: 'DynamoDB ConsumedWriteCapacityUnits'
                  alarm_actions: [ 'arn:aws:sns:us-east-1:1234:my-alarm' ]
                  insufficient_data_actions: []
                  ok_actions: [ 'arn:aws:sns:us-east-1:1234:my-alarm' ]
        - keyid: GKTADJGHEIQSXMKKRBJ08H
        - key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
        - region: us-east-1

You can also use alarms from pillars, and override values from the pillar
alarms by setting overrides on the resource. Note that 'boto_dynamodb_alarms'
will be used as a default value for all resources, if defined and can be
used to ensure alarms are always set for a resource.

Setting the alarms in a pillar:

.. code-block:: yaml

    boto_dynamodb_alarms:
      ConsumedWriteCapacityUnits:
        name: 'DynamoDB ConsumedWriteCapacityUnits **MANAGED BY SALT**'
        attributes:
          metric: ConsumedWriteCapacityUnits
          namespace: AWS/DynamoDB
          statistic: Sum
          comparison: '>='
          # threshold_percent is used to calculate the actual threshold
          # based on the provisioned capacity for the table.
          threshold_percent: 0.75
          period: 300
          evaluation_periods: 2
          unit: Count
          description: 'DynamoDB ConsumedWriteCapacityUnits'
          alarm_actions: [ 'arn:aws:sns:us-east-1:1234:my-alarm' ]
          insufficient_data_actions: []
          ok_actions: [ 'arn:aws:sns:us-east-1:1234:my-alarm' ]

    Ensure DynamoDB table exists:
      boto_dynamodb.present:
        - name: new_table
        - read_capacity_units: 1
        - write_capacity_units: 2
        - hash_key: primary_id
        - hash_key_data_type: N
        - range_key: start_timestamp
        - range_key_data_type: N
        - alarms:
             ConsumedWriteCapacityUnits:
                attributes:
                  threshold_percent: 0.90
                  period: 900
'''
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals
import datetime
import math
import sys
import logging
import copy

# Import salt libs
from salt.ext import six
import salt.utils.dictupdate as dictupdate

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s %(name)s %(levelname)s %(message)s',
    stream=sys.stdout
)
log = logging.getLogger()


class GsiNotUpdatableError(Exception):
    '''Raised when a global secondary index cannot be updated.'''
    pass


def __virtual__():
    '''
    Only load if boto_dynamodb is available.
    '''
    ret = 'boto_dynamodb' if 'boto_dynamodb.exists' in __salt__ else False
    return ret


def present(name=None,
            table_name=None,
            region=None,
            key=None,
            keyid=None,
            profile=None,
            read_capacity_units=None,
            write_capacity_units=None,
            alarms=None,
            alarms_from_pillar="boto_dynamodb_alarms",
            hash_key=None,
            hash_key_data_type=None,
            range_key=None,
            range_key_data_type=None,
            local_indexes=None,
            global_indexes=None,
            backup_configs_from_pillars='boto_dynamodb_backup_configs'):
    '''
    Ensure the DynamoDB table exists. Table throughput can be updated after
    table creation.

    Global secondary indexes (GSIs) are managed with some exceptions:

    - If a GSI deletion is detected, a failure will occur (deletes should be
      done manually in the AWS console).

    - If multiple GSIs are added in a single Salt call, a failure will occur
      (boto supports one creation at a time). Note that this only applies after
      table creation; multiple GSIs can be created during table creation.

    - Updates to existing GSIs are limited to read/write capacity only
      (DynamoDB limitation).

    name
        Name of the DynamoDB table

    table_name
        Name of the DynamoDB table (deprecated)

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string)
        that contains a dict with region, key and keyid.

    read_capacity_units
        The read throughput for this table

    write_capacity_units
        The write throughput for this table

    hash_key
        The name of the attribute that will be used as the hash key
        for this table

    hash_key_data_type
        The DynamoDB datatype of the hash key

    range_key
        The name of the attribute that will be used as the range key
        for this table

    range_key_data_type
        The DynamoDB datatype of the range key

    local_indexes
        The local indexes you would like to create

    global_indexes
        The global indexes you would like to create

    backup_configs_from_pillars
        Pillars to use to configure DataPipeline backups
    '''
    ret = {'name': name, 'result': True, 'comment': '', 'changes': {}}
    if table_name:
        ret['warnings'] = ['boto_dynamodb.present: `table_name` is deprecated.'
                           ' Please use `name` instead.']
        ret['name'] = table_name
        name = table_name

    comments = []
    changes_old = {}
    changes_new = {}

    # Ensure DynamoDB table exists
    table_exists = __salt__['boto_dynamodb.exists'](
        name,
        region,
        key,
        keyid,
        profile
    )
    if not table_exists:
        if __opts__['test']:
            ret['result'] = None
            ret['comment'] = 'DynamoDB table {0} would be created.'.format(name)
            return ret
        else:
            is_created = __salt__['boto_dynamodb.create_table'](
                name,
                region,
                key,
                keyid,
                profile,
                read_capacity_units,
                write_capacity_units,
                hash_key,
                hash_key_data_type,
                range_key,
                range_key_data_type,
                local_indexes,
                global_indexes
            )
            if not is_created:
                ret['result'] = False
                ret['comment'] = 'Failed to create table {0}'.format(name)
                _add_changes(ret, changes_old, changes_new)
                return ret

            comments.append('DynamoDB table {0} was successfully created'.format(name))
            changes_new['table'] = name
            changes_new['read_capacity_units'] = read_capacity_units
            changes_new['write_capacity_units'] = write_capacity_units
            changes_new['hash_key'] = hash_key
            changes_new['hash_key_data_type'] = hash_key_data_type
            changes_new['range_key'] = range_key
            changes_new['range_key_data_type'] = range_key_data_type
            changes_new['local_indexes'] = local_indexes
            changes_new['global_indexes'] = global_indexes
    else:
        comments.append('DynamoDB table {0} exists'.format(name))

    # Ensure DynamoDB table provisioned throughput matches
    description = __salt__['boto_dynamodb.describe'](
        name,
        region,
        key,
        keyid,
        profile
    )
    provisioned_throughput = description.get('Table', {}).get('ProvisionedThroughput', {})
    current_write_capacity_units = provisioned_throughput.get('WriteCapacityUnits')
    current_read_capacity_units = provisioned_throughput.get('ReadCapacityUnits')
    throughput_matches = (current_write_capacity_units == write_capacity_units and
                          current_read_capacity_units == read_capacity_units)
    if not throughput_matches:
        if __opts__['test']:
            ret['result'] = None
            comments.append('DynamoDB table {0} is set to be updated.'.format(name))
        else:
            is_updated = __salt__['boto_dynamodb.update'](
                name,
                throughput={
                    'read': read_capacity_units,
                    'write': write_capacity_units,
                },
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
            if not is_updated:
                ret['result'] = False
                ret['comment'] = 'Failed to update table {0}'.format(name)
                _add_changes(ret, changes_old, changes_new)
                return ret

            comments.append('DynamoDB table {0} was successfully updated'.format(name))
            changes_old['read_capacity_units'] = current_read_capacity_units,
            changes_old['write_capacity_units'] = current_write_capacity_units,
            changes_new['read_capacity_units'] = read_capacity_units,
            changes_new['write_capacity_units'] = write_capacity_units,
    else:
        comments.append('DynamoDB table {0} throughput matches'.format(name))

    provisioned_indexes = description.get('Table', {}).get('GlobalSecondaryIndexes', [])

    _ret = _global_indexes_present(provisioned_indexes, global_indexes, changes_old,
                                   changes_new, comments, name, region, key, keyid,
                                   profile)
    if not _ret['result']:
        comments.append(_ret['comment'])
        ret['result'] = _ret['result']
        if ret['result'] is False:
            ret['comment'] = ',\n'.join(comments)
            _add_changes(ret, changes_old, changes_new)
            return ret

    _ret = _alarms_present(name, alarms, alarms_from_pillar,
                           write_capacity_units, read_capacity_units,
                           region, key, keyid, profile)
    ret['changes'] = dictupdate.update(ret['changes'], _ret['changes'])
    comments.append(_ret['comment'])
    if not _ret['result']:
        ret['result'] = _ret['result']
        if ret['result'] is False:
            ret['comment'] = ',\n'.join(comments)
            _add_changes(ret, changes_old, changes_new)
            return ret

    # Ensure backup datapipeline is present
    datapipeline_configs = copy.deepcopy(
        __salt__['pillar.get'](backup_configs_from_pillars, [])
    )
    for config in datapipeline_configs:
        datapipeline_ret = _ensure_backup_datapipeline_present(
            name=name,
            schedule_name=config['name'],
            period=config['period'],
            utc_hour=config['utc_hour'],
            s3_base_location=config['s3_base_location'],
        )
        # Add comments and changes if successful changes were made (True for live mode,
        # None for test mode).
        if datapipeline_ret['result'] in [True, None]:
            ret['result'] = datapipeline_ret['result']
            comments.append(datapipeline_ret['comment'])
            if datapipeline_ret.get('changes'):
                ret['changes']['backup_datapipeline_{0}'.format(config['name'])] = \
                    datapipeline_ret.get('changes'),
        else:
            ret['comment'] = ',\n'.join([ret['comment'], datapipeline_ret['comment']])
            _add_changes(ret, changes_old, changes_new)
            return ret

    ret['comment'] = ',\n'.join(comments)
    _add_changes(ret, changes_old, changes_new)
    return ret


def _add_changes(ret, changes_old, changes_new):
    if changes_old:
        ret['changes']['old'] = changes_old
    if changes_new:
        ret['changes']['new'] = changes_new


def _global_indexes_present(provisioned_indexes, global_indexes, changes_old,
                            changes_new, comments, name, region, key, keyid,
                            profile):
    '''Handles global secondary index for the table present state.'''
    ret = {'result': True}
    if provisioned_indexes:
        provisioned_gsi_config = dict((index['IndexName'], index) for index in provisioned_indexes)
    else:
        provisioned_gsi_config = {}
    provisioned_index_names = set(provisioned_gsi_config.keys())

    # Map of index name to given Salt config for this run. This loop is complicated
    # because global_indexes is made up of OrderedDicts and lists.
    gsi_config = {}
    if global_indexes:
        for index in global_indexes:
            # Each index config is a key that maps to a list of OrderedDicts.
            index_config = index.values()[0]
            index_name = None
            for entry in index_config:
                # Key by the name field in the index config.
                if entry.keys() == ['name']:
                    index_name = entry.values()[0]
            if not index_name:
                ret['result'] = False
                ret['comment'] = 'Index name not found for table {0}'.format(name)
                return ret
            gsi_config[index_name] = index

    existing_index_names, new_index_names, index_names_to_be_deleted = _partition_index_names(
        provisioned_index_names, set(gsi_config.keys()))

    if index_names_to_be_deleted:
        ret['result'] = False
        ret['comment'] = ('Deletion of GSIs ({0}) is not supported! Please do this '
                          'manually in the AWS console.'.format(', '.join(index_names_to_be_deleted)))
        return ret
    elif len(new_index_names) > 1:
        ret['result'] = False
        ret['comment'] = ('Creation of multiple GSIs ({0}) is not supported due to API '
                          'limitations. Please create them one at a time.'.format(new_index_names))
        return ret

    if new_index_names:
        # Given the length check above, new_index_names should have a single element here.
        index_name = next(iter(new_index_names))
        _add_global_secondary_index(ret, name, index_name, changes_old, changes_new, comments,
                                    gsi_config, region, key, keyid, profile)
        if not ret['result']:
            return ret

    if existing_index_names:
        _update_global_secondary_indexes(ret, changes_old, changes_new, comments,
                                         existing_index_names, provisioned_gsi_config, gsi_config,
                                         name, region, key, keyid, profile)
        if not ret['result']:
            return ret

    if 'global_indexes' not in changes_old and 'global_indexes' not in changes_new:
        comments.append('All global secondary indexes match')
    return ret


def _partition_index_names(provisioned_index_names, index_names):
    '''Returns 3 disjoint sets of indexes: existing, to be created, and to be deleted.'''
    existing_index_names = set()
    new_index_names = set()
    for name in index_names:
        if name in provisioned_index_names:
            existing_index_names.add(name)
        else:
            new_index_names.add(name)
    index_names_to_be_deleted = provisioned_index_names - existing_index_names
    return existing_index_names, new_index_names, index_names_to_be_deleted


def _add_global_secondary_index(ret, name, index_name, changes_old, changes_new, comments,
                                gsi_config, region, key, keyid, profile):
    '''Updates ret iff there was a failure or in test mode.'''
    if __opts__['test']:
        ret['result'] = None
        ret['comment'] = 'Dynamo table {0} will have a GSI added: {1}'.format(
            name, index_name)
        return
    changes_new.setdefault('global_indexes', {})
    success = __salt__['boto_dynamodb.create_global_secondary_index'](
        name,
        __salt__['boto_dynamodb.extract_index'](
            gsi_config[index_name], global_index=True),
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )

    if success:
        comments.append('Created GSI {0}'.format(index_name))
        changes_new['global_indexes'][index_name] = gsi_config[index_name]
    else:
        ret['result'] = False
        ret['comment'] = 'Failed to create GSI {0}'.format(index_name)


def _update_global_secondary_indexes(ret, changes_old, changes_new, comments, existing_index_names,
                                     provisioned_gsi_config, gsi_config, name, region, key,
                                     keyid, profile):
    '''Updates ret iff there was a failure or in test mode.'''
    try:
        provisioned_throughputs, index_updates = _determine_gsi_updates(
            existing_index_names, provisioned_gsi_config, gsi_config)
    except GsiNotUpdatableError as e:
        ret['result'] = False
        ret['comment'] = six.text_type(e)
        return

    if index_updates:
        if __opts__['test']:
            ret['result'] = None
            ret['comment'] = 'Dynamo table {0} will have GSIs updated: {1}'.format(
                name, ', '.join(index_updates.keys()))
            return
        changes_old.setdefault('global_indexes', {})
        changes_new.setdefault('global_indexes', {})
        success = __salt__['boto_dynamodb.update_global_secondary_index'](
            name,
            index_updates,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        )

        if success:
            comments.append(
                'Updated GSIs with new throughputs {0}'.format(index_updates))
            for index_name in index_updates:
                changes_old['global_indexes'][index_name] = provisioned_throughputs[index_name]
                changes_new['global_indexes'][index_name] = index_updates[index_name]
        else:
            ret['result'] = False
            ret['comment'] = 'Failed to update GSI throughputs {0}'.format(index_updates)


def _determine_gsi_updates(existing_index_names, provisioned_gsi_config, gsi_config):
    # index name -> {'read': <read throughput>, 'write': <write throughput>}
    provisioned_throughputs = {}
    index_updates = {}
    for index_name in existing_index_names:
        current_config = provisioned_gsi_config[index_name]
        new_config = __salt__['boto_dynamodb.extract_index'](
            gsi_config[index_name], global_index=True).schema()

        # The provisioned config will have more fields than the new config, so only consider
        # fields in new_config.
        for key in new_config:
            if key in current_config and key != 'ProvisionedThroughput':
                new_value = new_config[key]
                current_value = current_config[key]
                # This is a special case since the Projection value can contain a list (not
                # correctly comparable with == as order doesn't matter).
                if key == 'Projection':
                    if new_value['ProjectionType'] != current_value['ProjectionType']:
                        raise GsiNotUpdatableError('GSI projection types do not match')
                    elif set(new_value.get('NonKeyAttributes', [])) != set(current_value.get('NonKeyAttributes', [])):
                        raise GsiNotUpdatableError('NonKeyAttributes do not match for GSI projection')
                elif new_value != current_value:
                    raise GsiNotUpdatableError(
                        'GSI property {0} cannot be updated for index {1}'.format(key, index_name))

        current_throughput = current_config.get('ProvisionedThroughput')
        current_read = current_throughput.get('ReadCapacityUnits')
        current_write = current_throughput.get('WriteCapacityUnits')
        provisioned_throughputs[index_name] = {'read': current_read, 'write': current_write}
        new_throughput = new_config.get('ProvisionedThroughput')
        new_read = new_throughput.get('ReadCapacityUnits')
        new_write = new_throughput.get('WriteCapacityUnits')
        if current_read != new_read or current_write != new_write:
            index_updates[index_name] = {'read': new_read, 'write': new_write}

    return provisioned_throughputs, index_updates


def _alarms_present(name, alarms, alarms_from_pillar,
                    write_capacity_units, read_capacity_units,
                    region, key, keyid, profile):
    '''helper method for present.  ensure that cloudwatch_alarms are set'''
    # load data from alarms_from_pillar
    tmp = copy.deepcopy(
        __salt__['config.option'](alarms_from_pillar, {})
    )
    # merge with data from alarms
    if alarms:
        tmp = dictupdate.update(tmp, alarms)
    # set alarms, using boto_cloudwatch_alarm.present
    merged_return_value = {'name': name, 'result': True, 'comment': '', 'changes': {}}
    for _, info in six.iteritems(tmp):
        # add dynamodb table to name and description
        info["name"] = name + " " + info["name"]
        info["attributes"]["description"] = name + " " + info["attributes"]["description"]
        # add dimension attribute
        info["attributes"]["dimensions"] = {"TableName": [name]}
        if info["attributes"]["metric"] == "ConsumedWriteCapacityUnits" \
           and "threshold" not in info["attributes"]:
            info["attributes"]["threshold"] = math.ceil(write_capacity_units * info["attributes"]["threshold_percent"])
            del info["attributes"]["threshold_percent"]
            # the write_capacity_units is given in unit / second. So we need
            # to multiply by the period to get the proper threshold.
            # http://docs.aws.amazon.com/amazondynamodb/latest/developerguide/MonitoringDynamoDB.html
            info["attributes"]["threshold"] *= info["attributes"]["period"]
        if info["attributes"]["metric"] == "ConsumedReadCapacityUnits" \
           and "threshold" not in info["attributes"]:
            info["attributes"]["threshold"] = math.ceil(read_capacity_units * info["attributes"]["threshold_percent"])
            del info["attributes"]["threshold_percent"]
            # the read_capacity_units is given in unit / second. So we need
            # to multiply by the period to get the proper threshold.
            # http://docs.aws.amazon.com/amazondynamodb/latest/developerguide/MonitoringDynamoDB.html
            info["attributes"]["threshold"] *= info["attributes"]["period"]
        # set alarm
        kwargs = {
            "name": info["name"],
            "attributes": info["attributes"],
            "region": region,
            "key": key,
            "keyid": keyid,
            "profile": profile,
        }
        results = __states__['boto_cloudwatch_alarm.present'](**kwargs)
        if not results["result"]:
            merged_return_value["result"] = results["result"]
        if results.get("changes", {}) != {}:
            merged_return_value["changes"][info["name"]] = results["changes"]
        if "comment" in results:
            merged_return_value["comment"] += results["comment"]
    return merged_return_value


def _ensure_backup_datapipeline_present(name, schedule_name, period,
                                        utc_hour, s3_base_location):

    kwargs = {
        'name': '{0}-{1}-backup'.format(name, schedule_name),
        'pipeline_objects': {
            'DefaultSchedule': {
                'name': schedule_name,
                'fields': {
                    'period': period,
                    'type': 'Schedule',
                    'startDateTime': _next_datetime_with_utc_hour(name, utc_hour).isoformat(),
                }
            },
        },
        'parameter_values': {
            'myDDBTableName': name,
            'myOutputS3Loc': '{0}/{1}/'.format(s3_base_location, name),
        }
    }
    return __states__['boto_datapipeline.present'](**kwargs)


def _get_deterministic_value_for_table_name(table_name, max_value):
    '''
    For a given table_name, returns hash of the table_name limited by max_value.
    '''
    return hash(table_name) % max_value


def _next_datetime_with_utc_hour(table_name, utc_hour):
    '''
    Datapipeline API is throttling us, as all the pipelines are started at the same time.
    We would like to uniformly distribute the startTime over a 60 minute window.

    Return the next future utc datetime where
        hour == utc_hour
        minute = A value between 0-59 (depending on table name)
        second = A value between 0-59 (depending on table name)
    '''
    today = datetime.date.today()

    # The minute and second values generated are deterministic, as we do not want
    # pipeline definition to change for every run.
    start_date_time = datetime.datetime(
        year=today.year,
        month=today.month,
        day=today.day,
        hour=utc_hour,
        minute=_get_deterministic_value_for_table_name(table_name, 60),
        second=_get_deterministic_value_for_table_name(table_name, 60)
    )

    if start_date_time < datetime.datetime.utcnow():
        one_day = datetime.timedelta(days=1)
        start_date_time += one_day

    return start_date_time


def absent(name,
           region=None,
           key=None,
           keyid=None,
           profile=None):
    '''
    Ensure the DynamoDB table does not exist.

    name
        Name of the DynamoDB table.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string)
        that contains a dict with region, key and keyid.
    '''
    ret = {'name': name, 'result': True, 'comment': '', 'changes': {}}
    exists = __salt__['boto_dynamodb.exists'](
        name,
        region,
        key,
        keyid,
        profile
    )
    if not exists:
        ret['comment'] = 'DynamoDB table {0} does not exist'.format(name)
        return ret

    if __opts__['test']:
        ret['comment'] = 'DynamoDB table {0} is set to be deleted'.format(name)
        ret['result'] = None
        return ret

    is_deleted = __salt__['boto_dynamodb.delete'](name, region, key, keyid, profile)
    if is_deleted:
        ret['comment'] = 'Deleted DynamoDB table {0}'.format(name)
        ret['changes'].setdefault('old', 'Table {0} exists'.format(name))
        ret['changes'].setdefault('new', 'Table {0} deleted'.format(name))
    else:
        ret['comment'] = 'Failed to delete DynamoDB table {0}'.format(name)
        ret['result'] = False
    return ret



# -*- coding: utf-8 -*-
'''
Management of Zabbix users.

:codeauthor: Jiri Kotlin <jiri.kotlin@ultimum.io>


'''

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals
from json import loads, dumps
from copy import deepcopy

# Import Salt libs
from salt.ext import six


def __virtual__():
    '''
    Only make these states available if Zabbix module is available.
    '''
    return 'zabbix.user_create' in __salt__


def present(alias, passwd, usrgrps, medias, password_reset=False, **kwargs):
    '''
    Ensures that the user exists, eventually creates new user.
    NOTE: use argument firstname instead of name to not mess values with name from salt sls.

    .. versionadded:: 2016.3.0

    :param alias: user alias
    :param passwd: user's password
    :param usrgrps: user groups to add the user to
    :param medias: user's medias to create
    :param password_reset: whether or not to reset password at update
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)
    :param firstname: string with firstname of the user, use 'firstname' instead of 'name' parameter to not mess \
    with value supplied from Salt sls file.

    .. code-block:: yaml

        make_user:
            zabbix_user.present:
                - alias: George
                - passwd: donottellanyonE@456x
                - password_reset: True
                - usrgrps:
                    - 13
                    - 7
                - medias:
                    - me@example.com:
                        - mediatype: mail
                        - period: '1-7,00:00-24:00'
                        - severity: NIWAHD
                    - make_jabber:
                        - active: true
                        - mediatype: jabber
                        - period: '1-5,08:00-19:00'
                        - sendto: jabbera@example.com
                    - text_me_morning_disabled:
                        - active: false
                        - mediatype: sms
                        - period: '1-5,09:30-10:00'
                        - severity: D
                        - sendto: '+42032132588568'

    '''
    connection_args = {}
    if '_connection_user' in kwargs:
        connection_args['_connection_user'] = kwargs['_connection_user']
    if '_connection_password' in kwargs:
        connection_args['_connection_password'] = kwargs['_connection_password']
    if '_connection_url' in kwargs:
        connection_args['_connection_url'] = kwargs['_connection_url']

    ret = {'name': alias, 'changes': {}, 'result': False, 'comment': ''}

    # Comment and change messages
    comment_user_created = 'User {0} created.'.format(alias)
    comment_user_updated = 'User {0} updated.'.format(alias)
    comment_user_notcreated = 'Unable to create user: {0}. '.format(alias)
    comment_user_exists = 'User {0} already exists.'.format(alias)
    changes_user_created = {alias: {'old': 'User {0} does not exist.'.format(alias),
                                    'new': 'User {0} created.'.format(alias),
                                    }
                            }

    def _media_format(medias_data):
        '''
        Formats medias from SLS file into valid JSON usable for zabbix API.
        Completes JSON with default values.

        :param medias_data: list of media data from SLS file

        '''
        if not medias_data:
            return list()
        medias_json = loads(dumps(medias_data))
        medias_attr = ('active', 'mediatype', 'period', 'severity', 'sendto')
        media_type = {'mail': 1, 'jabber': 2, 'sms': 3}
        media_severities = ('D', 'H', 'A', 'W', 'I', 'N')

        medias_dict = dict()
        for media in medias_json:
            for med in media:
                medias_dict[med] = dict()
                for medattr in media[med]:
                    for key, value in medattr.items():
                        if key in medias_attr:
                            medias_dict[med][key] = value

        medias_list = list()
        for key, value in medias_dict.items():
            # Load media values or default values
            active = '0' if six.text_type(value.get('active', 'true')).lower() == 'true' else '1'
            mediatype_sls = six.text_type(value.get('mediatype', 'mail')).lower()
            mediatypeid = six.text_type(media_type.get(mediatype_sls, 1))
            period = value.get('period', '1-7,00:00-24:00')
            sendto = value.get('sendto', key)

            severity_sls = value.get('severity', 'HD')
            severity_bin = six.text_type()
            for sev in media_severities:
                if sev in severity_sls:
                    severity_bin += '1'
                else:
                    severity_bin += '0'
            severity = six.text_type(int(severity_bin, 2))

            medias_list.append({'active': active,
                                'mediatypeid': mediatypeid,
                                'period': period,
                                'sendto': sendto,
                                'severity': severity})
        return medias_list

    user_exists = __salt__['zabbix.user_exists'](alias, **connection_args)

    if user_exists:
        user = __salt__['zabbix.user_get'](alias, **connection_args)[0]
        userid = user['userid']

        update_usrgrps = False
        update_medias = False

        usergroups = __salt__['zabbix.usergroup_get'](userids=userid, **connection_args)
        cur_usrgrps = list()

        for usergroup in usergroups:
            cur_usrgrps.append(int(usergroup['usrgrpid']))

        if set(cur_usrgrps) != set(usrgrps):
            update_usrgrps = True

        user_medias = __salt__['zabbix.user_getmedia'](userid, **connection_args)
        medias_formated = _media_format(medias)

        if user_medias:
            user_medias_copy = deepcopy(user_medias)
            for user_med in user_medias_copy:
                user_med.pop('userid')
                user_med.pop('mediaid')
            media_diff = [x for x in medias_formated if x not in user_medias_copy] + \
                         [y for y in user_medias_copy if y not in medias_formated]
            if media_diff:
                update_medias = True
        elif not user_medias and medias:
            update_medias = True

    # Dry run, test=true mode
    if __opts__['test']:
        if user_exists:
            if update_usrgrps or password_reset or update_medias:
                ret['result'] = None
                ret['comment'] = comment_user_updated
            else:
                ret['result'] = True
                ret['comment'] = comment_user_exists
        else:
            ret['result'] = None
            ret['comment'] = comment_user_created

    error = []

    if user_exists:
        ret['result'] = True
        if update_usrgrps or password_reset or update_medias:
            ret['comment'] = comment_user_updated

            if update_usrgrps:
                __salt__['zabbix.user_update'](userid, usrgrps=usrgrps, **connection_args)
                updated_groups = __salt__['zabbix.usergroup_get'](userids=userid, **connection_args)

                cur_usrgrps = list()
                for usergroup in updated_groups:
                    cur_usrgrps.append(int(usergroup['usrgrpid']))

                usrgrp_diff = list(set(usrgrps) - set(cur_usrgrps))

                if usrgrp_diff:
                    error.append('Unable to update grpup(s): {0}'.format(usrgrp_diff))

                ret['changes']['usrgrps'] = six.text_type(updated_groups)

            if password_reset:
                updated_password = __salt__['zabbix.user_update'](userid, passwd=passwd, **connection_args)
                if 'error' in updated_password:
                    error.append(updated_groups['error'])
                else:
                    ret['changes']['passwd'] = 'updated'

            if update_medias:
                for user_med in user_medias:
                    deletedmed = __salt__['zabbix.user_deletemedia'](user_med['mediaid'], **connection_args)
                    if 'error' in deletedmed:
                        error.append(deletedmed['error'])

                for media in medias_formated:
                    updatemed = __salt__['zabbix.user_addmedia'](userids=userid,
                                                                 active=media['active'],
                                                                 mediatypeid=media['mediatypeid'],
                                                                 period=media['period'],
                                                                 sendto=media['sendto'],
                                                                 severity=media['severity'],
                                                                 **connection_args)

                    if 'error' in updatemed:
                        error.append(updatemed['error'])

                ret['changes']['medias'] = six.text_type(medias_formated)

        else:
            ret['comment'] = comment_user_exists
    else:
        user_create = __salt__['zabbix.user_create'](alias, passwd, usrgrps, **kwargs)

        if 'error' not in user_create:
            ret['result'] = True
            ret['comment'] = comment_user_created
            ret['changes'] = changes_user_created
        else:
            ret['result'] = False
            ret['comment'] = comment_user_notcreated + six.text_type(user_create['error'])

    # error detected
    if error:
        ret['changes'] = {}
        ret['result'] = False
        ret['comment'] = six.text_type(error)

    return ret


def absent(name, **kwargs):
    '''
    Ensures that the user does not exist, eventually delete user.

    .. versionadded:: 2016.3.0

    :param name: user alias
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)

    .. code-block:: yaml

        George:
            zabbix_user.absent

    '''
    connection_args = {}
    if '_connection_user' in kwargs:
        connection_args['_connection_user'] = kwargs['_connection_user']
    if '_connection_password' in kwargs:
        connection_args['_connection_password'] = kwargs['_connection_password']
    if '_connection_url' in kwargs:
        connection_args['_connection_url'] = kwargs['_connection_url']

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    # Comment and change messages
    comment_user_deleted = 'USer {0} deleted.'.format(name)
    comment_user_notdeleted = 'Unable to delete user: {0}. '.format(name)
    comment_user_notexists = 'User {0} does not exist.'.format(name)
    changes_user_deleted = {name: {'old': 'User {0} exists.'.format(name),
                                   'new': 'User {0} deleted.'.format(name),
                                   }
                            }

    user_get = __salt__['zabbix.user_get'](name, **connection_args)

    # Dry run, test=true mode
    if __opts__['test']:
        if not user_get:
            ret['result'] = True
            ret['comment'] = comment_user_notexists
        else:
            ret['result'] = None
            ret['comment'] = comment_user_deleted
            ret['changes'] = changes_user_deleted

    if not user_get:
        ret['result'] = True
        ret['comment'] = comment_user_notexists
    else:
        try:
            userid = user_get[0]['userid']
            user_delete = __salt__['zabbix.user_delete'](userid, **connection_args)
        except KeyError:
            user_delete = False

        if user_delete and 'error' not in user_delete:
            ret['result'] = True
            ret['comment'] = comment_user_deleted
            ret['changes'] = changes_user_deleted
        else:
            ret['result'] = False
            ret['comment'] = comment_user_notdeleted + six.text_type(user_delete['error'])

    return ret



# -*- coding: utf-8 -*-
'''
Display output in a table format
=================================

.. versionadded:: 2017.7.0

This outputter displays a sequence of rows as table.

Example output::

    edge01.bjm01:
    ----------
        comment:
        ----------
        out:
        ----------
            ______________________________________________________________________________
            | Active | Interface | Last Move |        Mac        | Moves | Static | Vlan |
            ______________________________________________________________________________
            |  True  |  ae1.900  |    0.0    | 40:A6:77:5A:50:01 |   0   | False  | 111  |
            ______________________________________________________________________________
            |  True  |  ae1.111  |    0.0    | 64:16:8D:32:26:58 |   0   | False  | 111  |
            ______________________________________________________________________________
            |  True  |  ae1.111  |    0.0    | 8C:60:4F:73:2D:57 |   0   | False  | 111  |
            ______________________________________________________________________________
            |  True  |  ae1.111  |    0.0    | 8C:60:4F:73:2D:7C |   0   | False  | 111  |
            ______________________________________________________________________________
            |  True  |  ae1.222  |    0.0    | 8C:60:4F:73:2D:57 |   0   | False  | 222  |
            ______________________________________________________________________________
            |  True  |  ae1.222  |    0.0    | F4:0F:1B:76:9D:97 |   0   | False  | 222  |
            ______________________________________________________________________________
        result:
        ----------
'''

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

import operator
from functools import reduce  # pylint: disable=redefined-builtin

# Import Salt libs
import salt.output
import salt.utils.color
import salt.utils.data
import salt.utils.locales

# Import 3rd-party libs
from salt.ext import six
from salt.ext.six.moves import map, zip  # pylint: disable=redefined-builtin

__virtualname__ = 'table'


def __virtual__():
    return __virtualname__


class TableDisplay(object):
    '''
    Manage the table display content.
    '''

    _JUSTIFY_MAP = {
        'center': six.text_type.center,
        'right': six.text_type.rjust,
        'left': six.text_type.ljust
    }

    def __init__(self,
                 has_header=True,  # if header will be displayed
                 row_delimiter='-',  # row delimiter char
                 delim=' | ',  # column delimiter
                 justify='center',  # text justify
                 separate_rows=True,  # display the line separating two consecutive rows
                 prefix='| ',  # character to display at the beginning of the row
                 suffix=' |',  # character to display at the end of the row
                 width=50,  # column max width
                 wrapfunc=None):  # function wrapper
        self.__dict__.update(
            salt.utils.color.get_colors(
                __opts__.get('color'),
                __opts__.get('color_theme')
            )
        )
        self.strip_colors = __opts__.get('strip_colors', True)

        self.has_header = has_header
        self.row_delimiter = row_delimiter
        self.delim = delim
        self.justify = justify
        self.separate_rows = separate_rows
        self.prefix = prefix
        self.suffix = suffix
        self.width = width

        if not(wrapfunc and callable(wrapfunc)):
            self.wrapfunc = self.wrap_onspace
        else:
            self.wrapfunc = wrapfunc

    def ustring(self,
                indent,
                color,
                msg,
                prefix='',
                suffix='',
                endc=None):
        '''Build the unicode string to be displayed.'''
        if endc is None:
            endc = self.ENDC  # pylint: disable=no-member

        indent *= ' '
        fmt = u'{0}{1}{2}{3}{4}{5}'

        try:
            return fmt.format(indent, color, prefix, msg, endc, suffix)
        except UnicodeDecodeError:
            return fmt.format(indent, color, prefix, salt.utils.locales.sdecode(msg), endc, suffix)

    def wrap_onspace(self, text):

        '''
        When the text inside the column is longer then the width, will split by space and continue on the next line.'''

        def _truncate(line, word):
            return '{line}{part}{word}'.format(
                        line=line,
                        part=' \n'[(len(line[line.rfind('\n')+1:]) + len(word.split('\n', 1)[0]) >= self.width)],
                        word=word
                    )

        return reduce(_truncate, text.split(' '))

    def prepare_rows(self,
                     rows,
                     indent,
                     has_header):

        '''Prepare rows content to be displayed.'''

        out = []

        def row_wrapper(row):
            new_rows = [
                self.wrapfunc(item).split('\n')
                for item in row
            ]
            rows = []
            for item in map(lambda *args: args, *new_rows):
                if isinstance(item, (tuple, list)):
                    rows.append([substr or '' for substr in item])
                else:
                    rows.append([item])
            return rows

        logical_rows = [
            row_wrapper(row)
            for row in rows
        ]

        columns = map(lambda *args: args, *reduce(operator.add, logical_rows))

        max_widths = [
            max([len(six.text_type(item)) for item in column])
            for column in columns
        ]
        row_separator = self.row_delimiter * (len(self.prefix) + len(self.suffix) + sum(max_widths) +
                                              len(self.delim) * (len(max_widths) - 1))

        justify = self._JUSTIFY_MAP[self.justify.lower()]

        if self.separate_rows:
            out.append(
                self.ustring(
                    indent,
                    self.LIGHT_GRAY,  # pylint: disable=no-member
                    row_separator
                )
            )
        for physical_rows in logical_rows:
            for row in physical_rows:
                line = self.prefix \
                        + self.delim.join([
                                justify(six.text_type(item), width)
                                for (item, width) in zip(row, max_widths)
                        ]) + self.suffix
                out.append(
                    self.ustring(
                        indent,
                        self.WHITE,  # pylint: disable=no-member
                        line
                    )
                )
            if self.separate_rows or has_header:
                out.append(
                    self.ustring(
                        indent,
                        self.LIGHT_GRAY,  # pylint: disable=no-member
                        row_separator
                    )
                )
                has_header = False
        return out

    def display_rows(self,
                     rows,
                     labels,
                     indent):

        '''Prepares row content and displays.'''

        out = []

        if not rows:
            return out

        first_row_type = type(rows[0])
        # all rows must have the same datatype
        consistent = True
        for row in rows[1:]:
            if type(row) != first_row_type:
                consistent = False

        if not consistent:
            return out

        if isinstance(labels, dict):
            labels_temp = []
            for key in sorted(labels):
                labels_temp.append(labels[key])
            labels = labels_temp

        if first_row_type is dict:  # and all the others
            temp_rows = []
            if not labels:
                labels = [six.text_type(label).replace('_', ' ').title() for label in sorted(rows[0])]
            for row in rows:
                temp_row = []
                for key in sorted(row):
                    temp_row.append(six.text_type(row[key]))
                temp_rows.append(temp_row)
            rows = temp_rows
        elif isinstance(rows[0], six.string_types):
            rows = [[row] for row in rows]  # encapsulate each row in a single-element list

        labels_and_rows = [labels] + rows if labels else rows
        has_header = self.has_header and labels

        return self.prepare_rows(labels_and_rows, indent + 4, has_header)

    def display(self,
                ret,
                indent,
                out,
                rows_key=None,
                labels_key=None):

        '''Display table(s).'''

        rows = []
        labels = None

        if isinstance(ret, dict):
            if not rows_key or (rows_key and rows_key in list(ret.keys())):
                # either not looking for a specific key
                # either looking and found in the current root
                for key in sorted(ret):
                    if rows_key and key != rows_key:
                        continue  # if searching specifics, ignore anything else
                    val = ret[key]
                    if not rows_key:
                        out.append(
                            self.ustring(
                                indent,
                                self.DARK_GRAY,  # pylint: disable=no-member
                                key,
                                suffix=':'
                            )
                        )
                        out.append(
                            self.ustring(
                                indent,
                                self.DARK_GRAY,  # pylint: disable=no-member
                                '----------'
                            )
                        )
                    if isinstance(val, (list, tuple)):
                        rows = val
                        if labels_key:
                            # at the same depth
                            labels = ret.get(labels_key)  # if any
                        out.extend(self.display_rows(rows, labels, indent))
                    else:
                        self.display(val, indent + 4, out, rows_key=rows_key, labels_key=labels_key)
            elif rows_key:
                # dig deeper
                for key in sorted(ret):
                    val = ret[key]
                    self.display(val, indent, out, rows_key=rows_key, labels_key=labels_key)  # same indent
        elif isinstance(ret, (list, tuple)):
            if not rows_key:
                rows = ret
                out.extend(self.display_rows(rows, labels, indent))

        return out


def output(ret, **kwargs):
    '''
    Display the output as table.

    Args:

        * nested_indent: integer, specify the left alignment.
        * has_header: boolean specifying if header should be displayed. Default: True.
        * row_delimiter: character to separate rows. Default: ``_``.
        * delim: character to separate columns. Default: ``" | "``.
        * justify: text alignment. Default: ``center``.
        * separate_rows: boolean specifying if row separator will be displayed between consecutive rows. Default: True.
        * prefix: character at the beginning of the row. Default: ``"| "``.
        * suffix: character at the end of the row. Default: ``" |"``.
        * width: column max width. Default: ``50``.
        * rows_key: display the rows under a specific key.
        * labels_key: use the labels under a certain key. Otherwise will try to use the dictionary keys (if any).
        * title: display title when only one table is selected (using the ``rows_key`` argument).
    '''

    # to facilitate re-use
    if 'opts' in kwargs:
        global __opts__  # pylint: disable=W0601
        __opts__ = kwargs.pop('opts')

    # Prefer kwargs before opts
    base_indent = kwargs.get('nested_indent', 0) \
        or __opts__.get('out.table.nested_indent', 0)
    rows_key = kwargs.get('rows_key') \
        or __opts__.get('out.table.rows_key')
    labels_key = kwargs.get('labels_key') \
        or __opts__.get('out.table.labels_key')
    title = kwargs.get('title') \
        or __opts__.get('out.table.title')

    class_kvargs = {}
    argks = ('has_header', 'row_delimiter', 'delim', 'justify', 'separate_rows', 'prefix', 'suffix', 'width')

    for argk in argks:
        argv = kwargs.get(argk) \
            or __opts__.get('out.table.{key}'.format(key=argk))
        if argv is not None:
            class_kvargs[argk] = argv

    table = TableDisplay(**class_kvargs)

    out = []
    if title and rows_key:
        out.append(
            table.ustring(
                base_indent,
                title,
                table.WHITE,  # pylint: disable=no-member
                suffix='\n'
            )
        )

    return '\n'.join(table.display(salt.utils.data.decode(ret),
                                   base_indent,
                                   out,
                                   rows_key=rows_key,
                                   labels_key=labels_key))
